################################################################################
#################             Evaluate Module Tests            #################
################################################################################

import numpy as np
import unittest
from sklearn.metrics import r2_score, mean_squared_error, mean_absolute_error, explained_variance_score
import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)

from pySAR.evaluate import Evaluate

#Individual evaluation functions working with pySAR but test case approach erroneous
# @unittest.expectedFailure
class EvaluateTests(unittest.TestCase):

    def setUp(self):
        """ Create dummy data to test evaluate class. """
        self.a = np.random.ranf(size=100).reshape((-1,1))
        self.a_ = np.random.ranf(size=100).reshape((-1,1))
        self.b = np.random.ranf(size =80).reshape((-1,1))
        self.c = np.array([1,2,3,4,5,6,7,8,9,10]).reshape((-1,1))
        self.d = np.array([10,9,8,7,6,5,4,3,2,1]).reshape((-1,1))
        self.e = np.random.randint(20, size=10).reshape((-1,1))
        self.f = np.random.randint(20, size=10).reshape((-1,1))

    def test_evaluate(self):
        """ Test Evaluate class initialisation and its attributes. """
#1.)
        #assert that value error is raised when input parameters to class are not same shape
        with self.assertRaises(ValueError, msg='Value Error raised, input parameters are not of the same shape.'):
            fail_eval = Evaluate(self.a,self.b)

        eval = Evaluate(self.a,self.a_)
        metrics = ['r2', 'mse', 'rmse', 'mae', 'rpd', 'explained_var']
#2.)
        #assert metric names exist as instance variables in Evaluate class
        for met in range(0, len(metrics)):
            self.assertTrue(metrics[met] in list(eval.__dict__.keys()))
#3.)
        #testing Y_true variable in the Evaluate object has been set according to Y_true input parameter
        np_equality = np.testing.assert_array_equal(self.a, eval.Y_true)
        self.assertIsNone(np_equality)
#4.)
        #testing Y_pred variable in the Evaluate object has been set according to Y_pred input parameter
        np_equality = np.testing.assert_array_equal(self.a_, eval.Y_pred)
        self.assertIsNone(np_equality)
#5.)
        #numpy.testing function will raise an assertion error if the arrays are not equal
        with self.assertRaises(AssertionError, msg='Assertion Error raised, input arrays are not equal.'):
             np.testing.assert_array_equal(self.a_, eval.Y_true )
#6.)
        self.assertTrue(eval.Y_true.shape == (100,1))
        self.assertTrue(eval.Y_pred.shape == (100,1))

        #############################################
        eval = Evaluate(self.c,self.d)
        metrics = ['r2', 'mse', 'rmse', 'mae', 'rpd', 'explained_var']
#7.)
        #assert metric names exist as instance variables in Evaluate class
        for met in range(0, len(metrics)):
            self.assertTrue(metrics[met] in list(eval.__dict__.keys()))
#8.)
        np_equality = np.testing.assert_array_equal(self.c, eval.Y_true)
        self.assertIsNone(np_equality)
#9.)
        np_equality = np.testing.assert_array_equal(self.d, eval.Y_pred)
        self.assertIsNone(np_equality)
#10.)
        #numpy.testing function will raise an assertion error if the arrays are not equal
        with self.assertRaises(AssertionError, msg='Assertion Error raised, input arrays are not equal.'):
             np.testing.assert_array_equal(self.d, eval.Y_true )
#11.)
        self.assertTrue(eval.Y_true.shape == (10,1))
        self.assertTrue(eval.Y_pred.shape == (10,1))

#     @unittest.expectedFailure
    def test_r2(self):
        """ Test case for testing the R2 score method. """
#1.)
        with warnings.catch_warnings(record=True) as w: #catch Divide by zero runtime warning
            warnings.simplefilter("always")
            eval = Evaluate(self.a, self.a)
            self.assertEqual(len(w), 1)
            self.assertTrue("divide by zero encountered in double_scalars" in str(w[-1].message))

        #assert that R2 = 1 if Y_true and Y_pred arrays are equal
        self.assertEqual(eval.r2, 1)
        #assert r2 variable is of type float
        self.assertIsInstance(eval.r2, float)
#2.)
        #############################################
        eval = Evaluate(self.c, self.d)
        #assert r2 value is around the pre-calculated value for c=Y_true, d=Y_pred
        self.assertAlmostEqual(eval.r2, -3.0)
        #assert r2 variable is of type float
        self.assertIsInstance(eval.r2, float)
#3.)
        #############################################
        eval = Evaluate(self.e, self.f)
        actual_value = r2_score(self.e, self.f)
        self.assertAlmostEqual(eval.r2, actual_value)
        #assert r2 variable is of type float
        self.assertIsInstance(eval.r2, float)

    def test_mse(self):
        """ Test case for testing the MSE method. """
#1.)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            eval = Evaluate(self.a, self.a)
            self.assertEqual(len(w), 1)
            self.assertTrue("divide by zero encountered in double_scalars" in str(w[-1].message))

        self.assertEqual(eval.mse, 0)
        self.assertIsInstance(eval.mse, float)
#2.)
        #############################################
        eval = Evaluate(self.c, self.d)
        self.assertIsInstance(eval.mse, float)
#3.)
        #############################################
        eval = Evaluate(self.e, self.f)
        self.assertIsInstance(eval.mse, float)
        self.assertTrue(eval.mse>0)

    def test_rmse(self):
        """ Testing RMSE evaluation metric. """
#1.)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            eval = Evaluate(self.a, self.a)
            self.assertEqual(len(w), 1)
            self.assertTrue("divide by zero encountered in double_scalars" in str(w[-1].message))

        self.assertEqual(eval.rmse, 0)
        self.assertIsInstance(eval.rmse, float)
#2.)
        #############################################
        eval = Evaluate(self.c, self.d)
        self.assertIsInstance(eval.mse, float)
        self.assertTrue(eval.mse>0)
#3.)
        #############################################
        eval = Evaluate(self.e, self.f)
        self.assertIsInstance(eval.mse, float)
        self.assertTrue(eval.mse>0)

    def test_mae(self):
        """ Testing MAE evaluation metric. """
#1.)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            eval = Evaluate(self.a, self.a)
            self.assertEqual(len(w), 1)
            self.assertTrue("divide by zero encountered in double_scalars" in str(w[-1].message))

        self.assertEqual(eval.mae, 0)
        self.assertIsInstance(eval.mae, float)
#2.)
        #############################################
        eval = Evaluate(self.c, self.d)
        self.assertIsInstance(eval.mae, float)
        self.assertTrue(eval.mae>0)
#3.)
        #############################################
        eval = Evaluate(self.e, self.f)
        self.assertIsInstance(eval.mae, float)
        self.assertTrue(eval.mae>0)

    def test_explainedVar(self):
        """ Testing Explained Variance metric. """
#1.)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            eval = Evaluate(self.a, self.a)
            self.assertEqual(len(w), 1)
            self.assertTrue("divide by zero encountered in double_scalars" in str(w[-1].message))

        self.assertEqual(eval.explained_var, 1)
        self.assertIsInstance(eval.explained_var, float)
#2.)
        #############################################
        eval = Evaluate(self.c, self.d)
        self.assertIsInstance(eval.explained_var, float)
        self.assertTrue(eval.explained_var==-3)
#3.)
        #############################################
        eval = Evaluate(self.e, self.f)
        self.assertIsInstance(eval.explained_var, float)
        self.assertTrue(eval.explained_var<=1)

    def test_maxError(self):
        """ Testing max error evaluation metric. """
#1.)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            eval = Evaluate(self.a, self.a)
            self.assertEqual(len(w), 1)
            self.assertTrue("divide by zero encountered in double_scalars" in str(w[-1].message))

        self.assertEqual(eval.max_error_(), 0)
        self.assertIsInstance(eval.max_error_(), float)
#2.)
        #############################################
        eval = Evaluate(self.c, self.d)
        self.assertIsInstance(eval.max_error_(), float)
        self.assertTrue(eval.max_error_()==9)
#3.)
        #############################################
        eval = Evaluate(self.e, self.f)
        self.assertIsInstance(eval.max_error_(), float)
        self.assertTrue(eval.max_error_()>=1)

    @unittest.expectedFailure
    def test_meanPoissonDeviance(self):
        """ Testing mean poisson deviation metric. """
#1.)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            eval = Evaluate(self.a, self.a)
            self.assertEqual(len(w), 1)
            self.assertTrue("divide by zero encountered in double_scalars" in str(w[-1].message))

        self.assertEqual(eval.mean_poisson_deviance, 0)
        self.assertIsInstance(eval.mean_poisson_deviance, float)
#2.)
        #############################################
        eval = Evaluate(self.c, self.d)
        self.assertIsInstance(eval.mean_poisson_deviance, float)
        self.assertTrue(eval.mean_poisson_deviance>0)
#3.)
        #############################################
        eval = Evaluate(self.e, self.f)
        self.assertIsInstance(eval.mean_poisson_deviance, float)
        self.assertTrue(eval.mean_poisson_deviance>0)

    @unittest.expectedFailure
    def test_rpd(self):
        """ Testing RPD evaluation metric. """
#1.)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            eval = Evaluate(self.a, self.a)
            self.assertEqual(len(w), 1)
            self.assertTrue("divide by zero encountered in double_scalars" in str(w[-1].message))

        self.assertEqual(eval.rpd, 0)
        self.assertIsInstance(eval.rpd, float)
#2.)
        #############################################
        eval = Evaluate(self.c, self.d)
        self.assertIsInstance(eval.rpd, float)
        self.assertTrue(eval.rpd>0)
#3.)
        #############################################
        eval = Evaluate(self.e, self.f)
        self.assertIsInstance(eval.rpd, float)
        self.assertTrue(eval.rpd>0)




################################################################################
#################                    Model                     #################
################################################################################

from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR
from sklearn.linear_model import Lasso, LinearRegression, Ridge, SGDRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import RandomForestRegressor, AdaBoostRegressor, BaggingRegressor, GradientBoostingRegressor
from sklearn.cross_decomposition import PLSRegression
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.model_selection import GridSearchCV, train_test_split
from sklearn.metrics import SCORERS
from sklearn.exceptions import UndefinedMetricWarning
from sklearn.feature_selection import SelectKBest, chi2
from difflib import get_close_matches
import numpy as np
import os
import pickle

from .evaluate import Evaluate

class Model():
    """
    Class for building, fitting and training a various range of predictive
    regression models and all their related methods and attributes.

    Attributes
    ----------
    :algorithm : str
        sklearn regression algorithm to build and fit model with. Value can be 
        an approximate representation of model name, for example: 'plsreg' will 
        initialiase an instance of the PLSRegression model.
    :parameters : dict (default = {})
        parameters to use for building regression model, by default it is set to {}, meaning
        all of the models' default parameters are used.
    :test_split : float (default = 0.2)
        proportion of the test data to use for building model.

    Methods
    -------
    get_model():
        build model using inputtted parameters.
    train_test_split(X, Y, scale = True, test_size = 0.2,random_state=None, shuffle=True):
        get train-test split of dataset.
    fit():
        fit model.
    predict():
        predict activity values using model and test data.
    save(save_folder):
        save fitted model to save_folder.
    copy():
        make copy of model
    hyperparameter_tuning(parameters, metric='r2', cv=5):
        complete hyperparameter tuning of model and its associated parameters.
    model_fitted():
        return if model has been fitted (true or false)
    """
    def __init__(self, algorithm, parameters={}, test_split=0.2):

        self.algorithm = algorithm
        self.test_split = test_split

        #if no parameters input, then set to {} meaning default parameters are used
        if (parameters == [] or parameters == ""):
            self.parameters = {}
        else:
            self.parameters = parameters

        #list of valid models available to use for this class
        self.valid_models = ['plsregression','randomforestregressor','adaboostregressor',\
                            'baggingregressor','decisiontreeregressor', 'gbr', 'gradientboostingregressor', 
                            'linearregression', 'lasso', 'ridge', 'svr', 'sgd', 'kneighborsregressor', 'knn']

        #get closest match of valid model from the input algorithm parameter value
        model_matches = get_close_matches(self.algorithm.lower().strip(),[item.lower().strip() \
            for item in self.valid_models], cutoff=0.5)

        #if algorithm is a valid model then set it to self.algorithm, else raise error
        if model_matches!=[]:
            self.algorithm = model_matches[0]
        else:
            raise ValueError('Input algorithm {} not found in available valid models {}.'.format(
                    self.algorithm, self.valid_models))

        #create instance of algorithm object
        self.model = self.get_model()
        self.model_fit = None

    def get_model(self):
        """
        Create instance of model type specifed by input 'algorithm' argument. If
        input 'parameters' = {} then default parameters of sklearn model are used else set
        the parameters of the model to the values specified in the 'parameters' input.

        Parameters
        ----------
        :self (Model object) : 
            instance of Model class.

        Returns
        -------
        :model : sklearn.model
            instantiated regression model with default or user-specified parameters.
        """
        if self.algorithm.lower().strip() == 'plsregression':

            #get parameters of sklearn model and check that user inputted
            #  parameters are available in the model, only use those that are valid.
            model_params = set(dir(PLSRegression()))
            parameters = [i for i in model_params if i in self.parameters]

            #use default model parameters if user input parameters is empty, else
            #   use user-specified parameters.
            if parameters != {} or parameters !=[]:
                model = PLSRegression(**self.parameters)
            else:
                model = PLSRegression()

        elif self.algorithm.lower().strip() == 'randomforestregressor':

            model_params = set(dir(RandomForestRegressor()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {} or parameters !=[]:
                model = RandomForestRegressor(**self.parameters)
            else:
                model = RandomForestRegressor()

        elif self.algorithm.lower().strip() == 'adaboostregressor':

            model_params = set(dir(AdaBoostRegressor()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {} or parameters !=[]:
                model = AdaBoostRegressor(**self.parameters)
            else:
                model = AdaBoostRegressor()

        elif self.algorithm.lower().strip() == 'baggingregressor':

            model_params = set(dir(BaggingRegressor()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {} or parameters !=[]:
                model = BaggingRegressor(**self.parameters)
            else:
                model = BaggingRegressor()

        elif self.algorithm.lower().strip() == 'decisiontreeregressor':

            model_params = set(dir(DecisionTreeRegressor()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {} or parameters !=[]:
                model = DecisionTreeRegressor(**self.parameters)
            else:
                model = DecisionTreeRegressor()

        elif self.algorithm.lower().strip() == 'linearregression':

            model_params = set(dir(LinearRegression()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {} or parameters !=[]:
                model = LinearRegression(**self.parameters)
            else:
                model = LinearRegression()

        elif self.algorithm.lower().strip() == 'lasso':

            model_params = set(dir(Lasso()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {} or parameters !=[]:
                model = Lasso(**self.parameters)
            else:
                model = Lasso()

        elif self.algorithm.lower().strip() == 'ridge':

            model_params = set(dir(Ridge()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {} or parameters !=[]:
                model = Ridge(**self.parameters)
            else:
                model = Ridge()

        elif self.algorithm.lower().strip() == 'sgd':

            model_params = set(dir(SGDRegressor()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {} or parameters !=[]:
                model = SGDRegressor(**self.parameters)
            else:
                model = SGDRegressor()

        elif self.algorithm.lower().strip() == 'gbr' or \
            self.algorithm.lower().strip() == 'gradientboost' or \
            self.algorithm.lower().strip() == 'gradientboostingregressor':
            model_params = set(dir(GradientBoostingRegressor()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {} or parameters !=[]:
                model = GradientBoostingRegressor(**self.parameters)
            else:
                model = GradientBoostingRegressor()

        elif self.algorithm.lower().strip() == 'svr':

            model_params = set(dir(SVR()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {} or parameters !=[]:
                model = SVR(**self.parameters)
            else:
                model = SVR()

        elif self.algorithm.lower().strip() == 'knn' or \
           self.algorithm.lower().strip() == 'kneighborsregressor':

            model_params = set(dir(KNeighborsRegressor()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {} or parameters !=[]:
                model = KNeighborsRegressor(**self.parameters)
            else:
                model = KNeighborsRegressor()
        else:
            raise ValueError('Input Algorithm ({}) not found in available valid models'.
                format(self.algorithm, self.valid_models))

        return model

    def train_test_split(self, X, Y, scale=True, test_size=0.2, random_state=None, shuffle=True):
        """
        Split the X and Y input features and labels into random train and test
        subsets. By default a 80:20 split will be used, whereby 80% of the data
        will be used for training and 20% for testing. By default the input will
        be scaled first such that the mean is removed and features scaled to unit
        variance. By default data is shuffled before the split and random state is None.

        Parameters
        ----------
        :self (Model object)
            instance of Model class.
        :X : np.ndarray
            array of feaure data.
        :Y : np.ndarray
            array of observed label values.
        :scale : bool (default = True)
            if true then scale the features such that they are standardised.
        :test_size : float (default = 0.2)
            proportion of the total dataset to use for testing.
        :random_state : float (default = None)
            Controls the shuffling applied to the data before applying the split.
            Popular integer random seeds are 0 and 42.
        :shuffle : bool (default = True)
            Whether or not to shuffle the data before splitting.

        Returns
        -------
        self.X_train, self.X_test, self.Y_train, self.Y_test : np.ndarray
            splitted training and test data features and labels.
        """
        #validate that X and Y arrays are of the same size
        if (len(X)!=len(Y)):
            raise ValueError('X and Y input parameters must be of the same length, X: {}, Y: {}'.
                format(len(X), len(Y)))

        #reshape input arrays to 2D arrays
        if (X.ndim!=2):
            X = np.reshape(X, (-1,1))
        if (Y.ndim!=2):
            Y = np.reshape(Y, (-1,1))

        #if invalid test size input then set to default 0.2
        if (test_size <= 0 or test_size >=1):
            test_size = 0.2

        self.test_split = test_size     #setting test_size attribute

        #scale training data X
        if scale:
            X = StandardScaler().fit_transform(X)
            # X = MinMaxScaler().fit_transform(X)

        #split X and Y into training and test data
        X_train, X_test, Y_train, Y_test = train_test_split(X, Y,
            test_size=test_size, random_state=random_state, shuffle=shuffle)

        #set X and Y attributes
        self.X_train = X_train
        self.X_test = X_test
        self.Y_train = np.reshape(Y_train, (len(Y_train),))
        self.Y_test = np.reshape(Y_test, (len(Y_test),))

        return self.X_train, self.X_test, self.Y_train, self.Y_test

    def fit(self):
        """
        Fit model to training data and labels.

        Parameters
        ----------
        :self : (Model object) 
            instance of Model class.

        Returns
        -------
        :self.model_fit : np.ndarray
            fitted sklearn model of type specified by algorithm attribute.
        """
        self.model_fit = self.model.fit(self.X_train, self.Y_train)
        return self.model_fit

    def predict(self):
        """
        Predict the target values of unseen test data using the model.

        Parameters
        ----------
        :self : (Model object)
            instance of Model class.

        Returns
        -------
        :self.model_fit.predict(self.X_test) : np.ndarray
            array of predicted target values for unseen test data.
        """
        return self.model_fit.predict(self.X_test)

    def save(self, save_folder):
        """
        Save fitted model to specified save_folder.

        Parameters
        ----------
        :self : (Model object)
            instance of Model class.
        :save_folder : str
            folder to save model to.
        """
        save_path = os.path.join(save_folder, 'model.pkl')

        try:
            with open(save_path, 'wb') as file:
                pickle.dump(self.model, file)
        except (IOError, OSError, pickle.PickleError, pickle.UnpicklingError):
            print("Error pickling model with path: {} .".format(save_path))

    def hyperparameter_tuning(self, parameters={}, metric='r2', cv=5, n_jobs=None):
        """
        Hyperparamter tuning of model to find its optimal arrangment of parameters
        using a Grid Search.

        Parameters
        ----------
        :self : (Model object)
            instance of Model class.
        :parameters : dict (default = {})
            dictionary of parameter names and their values.
        :metric : str (default = r2)
            scoring metric used to evaluate the performance of the cross-validated
            model on the test set.
        :cv : int (default = 5)
            Determines the cross-validation splitting strategy.
        :n_jobs : int (default = None)
            Number of jobs to run in parallel. None means 1 job.

        Returns
        -------
        None
        """
        #input parameters parameter must be a dict, if not raise error
        if not isinstance(parameters, dict):
            raise TypeError('Parameters argument must be of type dict, not of type {}.'.format(type(parameters)))

        #input metric must be in available scoring metrics, if not raise error
        if metric not in sorted(SCORERS.keys()):
            raise UndefinedMetricWarning('Invalid scoring metric, \
                {} not in available Sklearn Scoring Metrics: {}\n'.format(
                    metric, SCORERS.keys()))

        #cv must be of type int and be between 5 and 10, if not then default of 5 is used
        if not (isinstance(cv, int)) or (cv<5 or cv>10):
            cv = 5

        #iterate through all parameter names to check if they are correct for model,
        #  if parameter not found in model params then delete from dictionary
        for p in list(parameters.keys()):
            if p not in (list(self.model.get_params().keys())):
                del parameters[p]

        #grid search of hyperparameter space for model
        model_copy = self.copy()
        grid_search = GridSearchCV(estimator=model_copy, param_grid=parameters, \
            cv=cv, scoring=metric, n_jobs=n_jobs, verbose=2, error_score=0)

        #fit X and Y to best model found in grid search
        grid_result = grid_search.fit(self.X_train, self.Y_train)

        #get best grid search metics values
        mean_test = grid_result.cv_results_['mean_test_score']
        std_test = grid_result.cv_results_['std_test_score']
        params = grid_result.cv_results_['params']

        #predict values of unseen test data using best found model
        best_model_pred = grid_result.predict(self.X_test)

        #create instance of Evaluate class and calculate metrics from best model
        eval = Evaluate(self.Y_test,best_model_pred)

        #print out results of grid search
        print('\n#############################################################')
        print('#################### Hyperparamter Results ####################')
        print('#############################################################\n')

        print('######################### Parameters ########################\n')
        print('# Best Params -> {}'.format(grid_result.best_params_))
        print('# Model Type -> {}'.format(repr(self)))
        print('# Scoring Metric -> {}'.format(metric))
        print('# CV -> {}'.format(cv))
        print('# Test Split -> {}'.format(self.test_split))

        print('########################## Metrics ###########################\n')
        print('# Best Score (R2) -> {}'.format(grid_result.best_score_))
        print('# RMSE: {} '.format(eval.rmse))
        print('# MSE: {} '.format(eval.mse))
        print('# MAE: {}'.format(eval.mae))
        print('# RPD {}'.format(eval.rpd))
        print('# Variance {}\n'.format(eval.explained_var))
        print('###############################################################')
        
    def copy(self):
        """
        Make a copy of the sklearn model stored in self.model instance variable.

        Parameters
        ----------
        :self : (Model object)
            instance of Model class.

        Returns
        -------
        :model_copy : sklearn.model
            deep copy of model.
        """
        model_copy = self.model
        return model_copy

    def model_fitted(self):
        """
        Return if model has been fitted, true or false.

        Parameters
        ----------
        :self : (Model object)
            instance of Model class.

        Returns
        -------
        :True/False : bool
            true if model (self.model) has been fitted, false if not.
        """
        return (self.model_fit != None)

    def feature_selection(self, X, Y, method=""):
        """
        Feature selection/dimensionality reduction on dataset and models.
        Return the best applicable features found using the technique selected
        from method input parameter.

        Parameters
        ----------
        :X : np.ndarray
            training data.
        :Y : np.ndarray
            training data labels.
        :method : str (default = "")
            feature selection method to use.

        Returns
        -------
        :X_new : np.ndarray
            best found features from training data.
        """
        valid_feature_selection = []

        #get closest valid feature selection method
        feature_matches = get_close_matches(method.lower().strip(), [item.lower().strip() \
            for item in valid_feature_selection], cutoff=0.5)

        #apply feature selection method according to input parameter
        if feature_matches == 'selectkbest':
            X_new = SelectKBest(chi2, k=2).fit_transform(X, Y)

        return X_new

        # https://scikit-learn.org/stable/modules/feature_selection.html
        
######################          Getters & Setters          ######################

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, val):
        self._model = val

    @property
    def test_split(self):
        return self._test_split

    @test_split.setter
    def test_split(self, val):
        self._test_split = val

    @property
    def valid_models(self):
        return self._valid_models

    @valid_models.setter
    def valid_models(self,val):
        self._valid_models = val

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self,val):
        self._parameters = val

    @property
    def algorithm(self):
        return self._algorithm

    @algorithm.setter
    def algorithm(self,val):
        self._algorithm = val

    @property
    def model_fit(self):
        return self._model_fit

    @model_fit.setter
    def model_fit(self,val):
        self._model_fit = val

    @property
    def valid_models(self):
        return self._valid_models

    @valid_models.setter
    def valid_models(self,val):
        self._valid_models = val

################################################################################

    def __str__(self):
        return "Model of type {} using {} parameters, model fit = {}".format(
            type(self.model).__name__, self.parameters, self.model_fitted())

    def __repr__(self):
        return type(self.model).__name__

    def __eq__(self, other):
        return self.model == other.model

    def __sizeof__(self):
        """ Get size of sklearn model """
        return self.model.__sizeof__()



################################################################################
#################                  Descriptors                 #################
################################################################################

import pandas as pd
import numpy as np
from difflib import get_close_matches
import json
from json import JSONDecodeError
import itertools
from tqdm import tqdm

from .globals_ import OUTPUT_DIR, OUTPUT_FOLDER, DATA_DIR
from .aaindex import  AAIndex
from .model import Model
from .pyDSP import PyDSP
from .evaluate import Evaluate
from .utils import *
from .descriptors.autocorrelation import moran_autocorrelation, geary_autocorrelation, norm_moreaubroto_autocorrelation
from .descriptors.composition import AAComposition, DipeptideComposition, TripeptideComposition, \
    sequenceOrderCorrelationFactor, pseudoAAC, amphiphilicPseudoAAC
from .descriptors.conjointTriad import conjoint_triad
from .descriptors.ctd import ctd_composition, ctd_transition, ctd_distribution, ctd_
from .descriptors.quasiSequenceOrder import quasi_sequence_order, seq_order_coupling_number

class Descriptors():
    """
    Class for calculating a wide variety of protein physiochemical and structural descriptors. 
    These descriptors have been used in a wide variety of Bioinformaitcs applications including:
    protein strucutral and functional class prediction, protein-protein interactions,
    subcellular location, secondary structure prediction etc. They represent the different
    structural, functional & interaction profiles of proteins by exploring the
    features in the groups of composition, correlation, distribution of the constituent
    residues and their biochemical and physiochemical properties.

    This class allows calculation of the following descriptors: amino acid compostion (AAComp),
    dipeptide composition (DPComp), tripeptide composition (TPComp), normalized Moreau-Broto
    autocorrelation (NMBAuto), Moran autocorrelation (MAuto), Geary autocorrelation (GAuto),
    Composition (C), Transition (T), Distribution (D), CTD, Conjoint Triad (CTriad), sequence
    order coupling number (SOCNum), Quasi-sequence-order (QSOrder), Pseudo amino-acid
    composition - type 1 (PAAcomp) and Amphiphilic pseudo amino-acid composition - type 2 (APAAComp). 
    The descriptors directory contains the individual modules and functions used to calculate
    these descriptors. 

    Similar to other classes in pySAR, this class works via configuration files which contain
    the values for all the potential parameters of each descriptor. An array of protein sequences 
    can be passed into the class, but if none are then they will be imported from the dataset using
    the dataset parameters in the config file. By default, the class will look for a descriptors csv
    which is a file of the pre-calcualted descriptor values for the specified dataset, if this file 
    doesn't exist, or the parameter value is blank, then each descriptor will be calculated using 
    its respective function, depending on whether its respective parameter is set in the config.

    It is reccomended that with every new dataset, the Descriptors class should be instantiated 
    with the all_desc value set to 1 in the config file. This will calculate all the descriptor
    values for the dataset of protein sequences, storing the result in a csv file, meaning that
    this file can be used for future use. 

    Attributes
    ----------
    :desc_config: str
        path to configuration file which will contain the various parameter values for all
        descriptors. If invalid value input then error will be raised.
    :protein_seqs : np.ndarray
        array of protein sequences that descriptors will be calculated for. If set to none, 
        protein sequences will be imported from the specified dataset in the config file.

    References
    ----------
    [1]: Dong, J., Yao, ZJ., Zhang, L. et al. PyBioMed: a python library for
         various molecular representations of chemicals, proteins and DNAs and
         their interactions. J Cheminform 10, 16 (2018).
         https://doi.org/10.1186/s13321-018-0270-2
    [2]: Reczko, M. and Bohr, H. (1994) The DEF data base of sequence based protein
         fold class predictions. Nucleic Acids Res, 22, 3616-3619.
    [3]: Hua, S. and Sun, Z. (2001) Support vector machine approach for protein
         subcellular localization prediction. Bioinformatics, 17, 721-728.
    [4]: Broto P, Moreau G, Vandicke C: Molecular structures: perception,
         autocorrelation descriptor and SAR studies. Eur J Med Chem 1984, 19: 71–78.
    [5]: Ong, S.A., Lin, H.H., Chen, Y.Z. et al. Efficacy of different protein
         descriptors in predicting protein functional families. BMC Bioinformatics
         8, 300 (2007). https://doi.org/10.1186/1471-2105-8-300
    [6]: Inna Dubchak, Ilya Muchink, Stephen R.Holbrook and Sung-Hou Kim. Prediction
         of protein folding class using global description of amino acid sequence.
         Proc.Natl. Acad.Sci.USA, 1995, 92, 8700-8704.
    [7]: Juwen Shen, Jian Zhang, Xiaomin Luo, Weiliang Zhu, Kunqian Yu, Kaixian Chen,
         Yixue Li, Huanliang Jiang. Predicting proten-protein interactions based only
         on sequences inforamtion. PNAS. 2007 (104) 4337-4341.
    [8]: Kuo-Chen Chou. Prediction of Protein Subcellar Locations by Incorporating
         Quasi-Sequence-Order Effect. Biochemical and Biophysical Research
         Communications 2000, 278, 477-483.<- quasi-seq-order refernece
    [9]: Kuo-Chen Chou. Prediction of Protein Cellular Attributes Using
         Pseudo-Amino Acid Composition. PROTEINS: Structure, Function, and
         Genetics, 2001, 43: 246-255.
    [10]: Kuo-Chen Chou. Using amphiphilic pseudo amino acid composition to predict enzyme
          subfamily classes. Bioinformatics, 2005,21,10-19.
    """
    def __init__(self, desc_config="", protein_seqs=None):

        self.desc_config = desc_config
        self.protein_seqs = protein_seqs
        self.parameters = {}

        desc_config_filepath = ""
        #open json config file
        if not isinstance(desc_config, str) or desc_config is None:
            raise TypeError('JSON config file must be a filepath of type string, got type {}'.format(type(desc_config)))
        if os.path.isfile(self.desc_config):
            desc_config_filepath = self.desc_config
        elif os.path.isfile(os.path.join('config', self.desc_config)):
            desc_config_filepath = os.path.join('config', self.desc_config)
        else:
            raise OSError('JSON config file not found at path: {}.'.format(desc_config_filepath))
        try:
            with open(desc_config_filepath) as f:
                self.parameters = json.load(f)
        except:
            raise JSONDecodeError('Error parsing config JSON file: {}.'.format(desc_config_filepath))

        #set descriptor parameters
        self.dataset_parameters = self.parameters["dataset"][0]
        self.descr_config = self.parameters["descriptors"]
        self.descr_parameters = self.parameters["descriptor_parameters"]
        self.all_desc = self.descr_config[0]["descriptors"]["all_desc"]
        
        if not isinstance(self.protein_seqs, pd.Series):
            if (self.protein_seqs is None or self.protein_seqs == ""): 
                dataset_filepath = ""
                #open dataset and read protein seqs if protein_seqs is empty/None
                if os.path.isfile(self.dataset_parameters["dataset"]):
                    dataset_filepath = self.dataset_parameters["dataset"]
                elif os.path.isfile(os.path.join(DATA_DIR, self.dataset_parameters["dataset"])):
                    dataset_filepath = os.path.join(DATA_DIR, self.dataset_parameters["dataset"])
                else:
                    raise OSError('Dataset file not found at path: {}.'.format(dataset_filepath))

                #read in dataset csv
                try:
                    data = pd.read_csv(dataset_filepath, sep=",", header=0)
                    self.protein_seqs = data[self.dataset_parameters["sequence_col"]]
                except:
                    raise IOError('Error opening dataset file: {}'.format(dataset_filepath))
            else: 
                #if 1 protein sequence (1 string) input then convert to pandas Series object
                if isinstance(self.protein_seqs, str):
                    self.protein_seqs = pd.Series(self.protein_seqs)

                #only the sequences should be passed in, not all columns in a dataset etc.
                if isinstance(self.protein_seqs, pd.DataFrame) and \
                    len(self.protein_seqs.columns) > 1:
                    raise ValueError('The full dataset must not be passed in, only the \
                        columns containing the protein sequences.')

        #remove any gaps from protein sequences
        self.protein_seqs = remove_gaps(self.protein_seqs)

        #validate that all input protein sequences are valid and only contain
        #  valid amino acids, if not then raise ValueError
        invalid_seqs = valid_sequence(self.protein_seqs)
        if invalid_seqs!=None:
            raise ValueError('Invalid Amino Acids found in protein sequence dataset: {}'.
                format(invalid_seqs))

        #get the total number of inputted protein sequences
        self.num_seqs = len(self.protein_seqs)

        #initialise all descriptor attributes to empty dataframes
        self.aa_composition = pd.DataFrame()
        self.dipeptide_composition = pd.DataFrame()
        self.tripeptide_composition = pd.DataFrame()
        self.normalized_moreaubroto_autocorrelation = pd.DataFrame()
        self.moran_autocorrelation = pd.DataFrame()
        self.geary_autocorrelation = pd.DataFrame()
        self.ctd = pd.DataFrame()
        self.comp = pd.DataFrame()
        self.transition = pd.DataFrame()
        self.distribution = pd.DataFrame()
        self.conjoint_triad = pd.DataFrame()
        self.seq_order_coupling_number = pd.DataFrame()
        self.quasi_seq_order = pd.DataFrame()
        self.pseudo_aa_composition = pd.DataFrame()
        self.amp_pseudo_aa_composition = pd.DataFrame()
        self.all_descriptors = pd.DataFrame()

        #try importing descriptors csv with pre-calculated descriptor values,
        #  if not found then calculate all descriptors if all_desc is true
        if os.path.isfile((os.path.join(DATA_DIR, self.descr_config[0]["descriptors_csv"]))):
            self.import_descriptors()
            #get the total number of inputted protein sequences
            self.num_seqs = self.all_descriptors.shape[0]
        else:
            #if all_desc parameter true then calculate all descriptor values and store in their respective attributes
            if (self.all_desc):
                self.all_descriptors = self.get_all_descriptors()
                #save all calculated descriptor values for next time
                self.all_descriptors.to_csv(os.path.join(DATA_DIR, self.descr_config[0]["descriptors_csv"]), index=0)

        #create dictionary of descriptors and their associated groups
        keys = self.all_descriptors_list()
        values = ["Composition"]*3 + ["Autocorrelation"]*3 + ["CTD"]*4 + ["Conjoint Triad"] + \
            ["Quasi-Sequence-Order"]*2 + ["Pseudo Composition"]*2
        self.descriptor_groups = dict(zip(keys,values))

        #get shape of descriptors
        self.shape = self.all_descriptors.shape

    def import_descriptors(self):
        """
        By default, the class will search for a file in the DATA_DIR called
        whatever the "descriptors_csv" parameter in the config file is set to. This
        file contains ALL of the pre-calculated descriptor values.
        This function parses this csv file and sets the descriptor instance variables
        to the correct feature values. The descriptors file should contain the values
        of all possible 15 descriptor parameters for the N protein sequences.
        To create a descriptors file, create an instance of the desc class,
        with the parameter all_desc=True in the config file, this will calculate
        all descriptors and store them to the DATA_DIR in a csv. The aim of this 
        function is to simply import the descriptors during each use of pySAR rather than
        calculating each time.

        Parameters
        ----------
        None
        """
        descr_filepath = os.path.join(DATA_DIR, self.descr_config[0]["descriptors_csv"])

        #verify descriptors csv exists
        if not (os.path.isfile(descr_filepath)):
            raise OSError('Descriptors csv file does not exist, at filepath: {}'.format(descr_filepath))

        #import descriptors csv as dataframe
        try:
            descriptor_df = pd.read_csv(descr_filepath)
        except IOError:
            print('Error reading descriptors csv file: {}.'.format(descr_filepath))

        #replacing any +/- infinity or NAN values with 0
        descriptor_df.replace([np.inf, -np.inf], np.nan)
        descriptor_df = descriptor_df.fillna(0)

        #calculate dimension of each descriptor in the csv
        #pull each descriptor value from the csv according to its dimension, setting
        #the values to the class instance variables
        aa_composition_dim = (0,20)
        self.aa_composition = descriptor_df.iloc[:,aa_composition_dim[0]:aa_composition_dim[1]]

        dipeptide_composition_dim = (20,420)
        self.dipeptide_composition = descriptor_df.iloc[:,dipeptide_composition_dim[0]:dipeptide_composition_dim[1]]

        tripeptide_composition_dim = (420,8420)
        self.tripeptide_composition = descriptor_df.iloc[:,tripeptide_composition_dim[0]:tripeptide_composition_dim[1]]

        #dimension of autocorrelation descriptors depends on the max lag value and number of properties
        norm_moreaubroto_dim = (8420,
            8420 + (self.descr_parameters[0]["normalized_moreaubroto_autocorrelation"][0]["lag"]*len(self.descr_parameters[0]["normalized_moreaubroto_autocorrelation"][0]["properties"])))
        self.normalized_moreaubroto_autocorrelation = descriptor_df.iloc[:,norm_moreaubroto_dim[0]:norm_moreaubroto_dim[1]]

        moran_auto_dim = (norm_moreaubroto_dim[1], norm_moreaubroto_dim[1] +
            (self.descr_parameters[0]["moran_autocorrelation"][0]["lag"]*len(self.descr_parameters[0]["moran_autocorrelation"][0]["properties"])))
        self.moran_autocorrelation = descriptor_df.iloc[:,moran_auto_dim[0]: moran_auto_dim[1]]

        geary_auto_dim = (moran_auto_dim[1], moran_auto_dim[1] +
            (self.descr_parameters[0]["geary_autocorrelation"][0]["lag"]*len(self.descr_parameters[0]["geary_autocorrelation"][0]["properties"])))
        self.geary_autocorrelation = descriptor_df.iloc[:,geary_auto_dim[0]:geary_auto_dim[1]]

        ctd_dim = (geary_auto_dim[1], geary_auto_dim[1]+147)
        self.ctd =  descriptor_df.iloc[:,ctd_dim[0]:ctd_dim[1]]

        composition_dim = (ctd_dim[1], ctd_dim[1]+21)
        self.comp = descriptor_df.iloc[:,composition_dim[0]:composition_dim[1]]
    
        transition_dim = (composition_dim[1], composition_dim[1]+21)
        self.transition = descriptor_df.iloc[:,transition_dim[0]:transition_dim[1]]

        distribution_dim = (transition_dim[1], transition_dim[1]+105)
        self.distribution = descriptor_df.iloc[:,distribution_dim[0]:distribution_dim[1]]

        conjoint_triad_dim = (distribution_dim[1], distribution_dim[1]+343)
        self.conjoint_triad = descriptor_df.iloc[:,conjoint_triad_dim[0]:conjoint_triad_dim[1]]

        #dimension of SOCNum depends on the maximum lag value 
        seq_order_coupling_number_dim = (conjoint_triad_dim[1],
            conjoint_triad_dim[1] + (self.descr_parameters[0]["seq_order_coupling_number"][0]["lag"])*2)
        self.seq_order_coupling_number = descriptor_df.iloc[:,seq_order_coupling_number_dim[0]:seq_order_coupling_number_dim[1]]

        quasi_seq_order_dim = (seq_order_coupling_number_dim[1], seq_order_coupling_number_dim[1] + 100)
        self.quasi_seq_order = descriptor_df.iloc[:,quasi_seq_order_dim[0]:quasi_seq_order_dim[1]]

        pseudo_aa_composition_dim = (quasi_seq_order_dim[1], quasi_seq_order_dim[1] + 50)
        self.pseudo_aa_composition = descriptor_df.iloc[:,pseudo_aa_composition_dim[0]:pseudo_aa_composition_dim[1]]

        amp_pseudo_aa_composition_dim = (pseudo_aa_composition_dim[1], pseudo_aa_composition_dim[1] + 80)
        self.amp_pseudo_aa_composition = descriptor_df.iloc[:,amp_pseudo_aa_composition_dim[0]:amp_pseudo_aa_composition_dim[1]]

        self.all_descriptors = descriptor_df.iloc[:,:]

        #**need to fix pseudo_aa_composition and amp_pseudo_aa_composition dims

    def get_aa_composition(self):
        """
        Calculate Amino Acid Composition (AAComp) of protein sequences using the
        respective function in the composition.py module in the descriptors directory.

        Returns
        -------
        :aa_composition : pd.Dataframe
            pandas dataframe of AAComp for protein sequence. Dataframe will
            be of the shape N x 20, where 20 is the number of features 
            calculated from the descriptor (for the 20 amino acids).
        """
        print('\nGetting AA Composition Descriptors...')
        print('#####################################\n')

        #if attribute already calculated & not empty then return it
        if not self.aa_composition.empty:
            return self.aa_composition

        #initialise dataframe
        aa_comp_df = pd.DataFrame()

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            aa_comp_seq = AAComposition(seq)
            aa_comp_df = pd.concat([aa_comp_df, aa_comp_seq])

        self.aa_composition = aa_comp_df

        return self.aa_composition

    def get_dipeptide_composition(self):
        """
        Calculate Dipeptide Composition (DPComp) of protein sequences using the
        respective function in the composition.py module in the descriptors directory.

        Returns
        -------
        :dipeptide_composition : pd.Dataframe
            pandas Dataframe of dipeptide composition for protein sequence. Dataframe will
            be of the shape N x 400, where 400 is the number of features calculated 
            from the descriptor (20^2 for the 20 canonical amino acids).
        """
        print('\nGetting Dipeptide Composition Descriptors...')
        print('############################################\n')

        #if attribute already calculated & not empty then return it
        if not self.dipeptide_composition.empty:
            return self.dipeptide_composition

        #initialise dataframe
        dipeptide_comp_df = pd.DataFrame()
        
        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            dipeptide_comp_seq = DipeptideComposition(seq)
            dipeptide_comp_df = pd.concat([dipeptide_comp_df, dipeptide_comp_seq])

        self.dipeptide_composition = dipeptide_comp_df

        return self.dipeptide_composition

    def get_tripeptide_composition(self):
        """ 
        Calculate Tripeptide Composition (TPComp) of protein sequences using the
        respective function in the composition.py module in the descriptors directory.

        Returns
        -------
        :tripeptide_composition : pd.Dataframe
            pandas Dataframe of tripeptide composition for protein sequence. Dataframe will
            be of the shape N x 8000, where 8000 is the number of features calculated 
            from the descriptor (20^3 for the 20 canonical amino acids).
        """
        print('\nGetting Tripeptide Composition Descriptors...')
        print('#############################################\n')

        #if attribute already calculated & not empty then return it
        if not self.tripeptide_composition.empty:
            return self.tripeptide_composition

        #initialise dataframe
        tripeptide_comp_df = pd.DataFrame()
        
        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            tripeptide_comp_seq = TripeptideComposition(seq)
            tripeptide_comp_df = pd.concat([tripeptide_comp_df, tripeptide_comp_seq])

        self.tripeptide_composition = tripeptide_comp_df

        return self.tripeptide_composition

    def get_norm_moreaubroto_autocorrelation(self):
        """
        Calculate Normalized Moreau-Broto Autocorrelation (NMBAuto) of protein sequences using the
        respective function in the autocorrelation.py module in the descriptors directory.

        Returns
        -------
        :normalized_moreaubroto_autocorrelation : pd.Dataframe
            pandas Dataframe of NMBAuto values for protein sequence. Output will
            be of the shape N x M, where M is the number of features calculated from
            the descriptor. By default, the shape will be N x 240 (30 features per 
            property - using 8 properties).
        """
        print('\nGetting Normalized Moreaubroto Autocorrelation Descriptors...')
        print('#############################################################\n')

        #if attribute already calculated & not empty then return it
        if not self.normalized_moreaubroto_autocorrelation.empty:
            return self.normalized_moreaubroto_autocorrelation

        #get descriptor-specific parameters from config file
        lag = self.descr_parameters[0]["normalized_moreaubroto_autocorrelation"][0]["lag"]
        properties = self.descr_parameters[0]["normalized_moreaubroto_autocorrelation"][0]["properties"]

        #initialise dataframe
        norm_moreaubroto_df = pd.DataFrame()

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            norm_moreaubroto_seq = norm_moreaubroto_autocorrelation(seq, lag=lag, properties=properties)
            norm_moreaubroto_df = pd.concat([norm_moreaubroto_df, norm_moreaubroto_seq])
            
        self.normalized_moreaubroto_autocorrelation = norm_moreaubroto_df

        return self.normalized_moreaubroto_autocorrelation

    def get_moran_autocorrelation(self):
        """
        Calculate Moran Autocorrelation (MAuto) of protein sequences using the
        respective function in the autocorrelation.py module in the descriptors directory.

        Returns
        -------
        :moran_autocorrelation : pd.DataFrame
            pandas Dataframe of MAuto values for protein sequence. Output will
            be of the shape N x M, where M is the number of features calculated from
            the descriptor. By default, the shape will be N x 240 (30 features per 
            property - using 8 properties).
        """
        print('\nGetting Moran Autocorrelation Descriptors...')
        print('############################################\n')

        #if attribute already calculated & not empty then return it
        if not self.moran_autocorrelation.empty:
            return self.moran_autocorrelation

        #get descriptor-specific parameters from config file
        lag = self.descr_parameters[0]["moran_autocorrelation"][0]["lag"]
        properties = self.descr_parameters[0]["moran_autocorrelation"][0]["properties"]

        #initialise dataframe
        moran_df = pd.DataFrame()

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            moran_seq = moran_autocorrelation(seq, lag=lag, properties=properties)
            moran_df = pd.concat([moran_df, moran_seq])

        self.moran_autocorrelation = moran_df

        return self.moran_autocorrelation 
        
    def get_geary_autocorrelation(self):
        """
        Calculate Geary Autocorrelation (GAuto) of protein sequences using the
        respective function in the autocorrelation.py module in the descriptors directory.

        Returns
        -------
        :geary_autocorrelation : pd.DataFrame
            pandas Dataframe of GAuto values for protein sequence. Output will
            be of the shape N x M, where M is the number of features calculated from
            the descriptor. By default, the shape will be N x 240 (30 features per 
            property - using 8 properties).
        """
        print('\nGetting Geary Autocorrelation Descriptors...')
        print('############################################\n')

        #if attribute already calculated & not empty then return it
        if not self.geary_autocorrelation.empty:
            return self.geary_autocorrelation

        #get descriptor-specific parameters from config file
        lag = self.descr_parameters[0]["geary_autocorrelation"][0]["lag"]
        properties = self.descr_parameters[0]["geary_autocorrelation"][0]["properties"]

        #initialise dataframe
        geary_df = pd.DataFrame()

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            geary_seq = geary_autocorrelation(seq, lag=lag, properties=properties)
            geary_df = pd.concat([geary_df, geary_seq])

        self.geary_autocorrelation = geary_df

        return self.geary_autocorrelation 

    def get_composition(self):
        """ 
        Calculate Composition (C_CTD) of protein sequences using the
        respective function in the ctd.py module in the descriptors directory.

        Returns
        -------
        :composition : pd.DataFrame
            pandas dataframe of C_CTD values for protein sequence. Output will
            be of the shape N x 3, where 3 is the number of features calculated from
            the descriptor.
        """
        print('\nGetting Composition (CTD) Descriptors...')
        print('#########################################\n')

        #if attribute already calculated & not empty then return it
        if not self.comp.empty:
            return self.comp

        #get descriptor-specific parameters from config file
        property = self.descr_parameters[0]["composition"][0]["property"]

        #initialise dataframe
        comp_df = pd.DataFrame()

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            comp_seq = ctd_composition(seq, property=property)
            comp_df = pd.concat([comp_df, comp_seq])

        self.comp = comp_df

        return self.comp

    def get_transition(self):
        """ 
        Calculate Transition (T_CTD) of protein sequences using the
        respective function in the ctd.py module in the descriptors directory.

        Returns
        -------
        :transition : pd.Dataframe
            pandas Dataframe of T_CTD values for protein sequence. Output will
            be of the shape N x 1, where 3 is the number of features calculated from
            the descriptor.
        """
        print('\nGetting Transition (CTD) Descriptors...')
        print('#########################################\n')

        #if attribute already calculated & not empty then return it
        if not self.transition.empty:
            return self.transition

        #get descriptor-specific parameters from config file
        property = self.descr_parameters[0]["transition"][0]["property"]

        #initialise dataframe
        transition_df = pd.DataFrame()

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            transition_seq = ctd_transition(seq, property=property)
            transition_df = pd.concat([transition_df, transition_seq])

        self.transition = transition_df

        return self.transition

    def get_distribution(self):
        """ 
        Calculate Distribution (D_CTD) of protein sequences using the
        respective function in the ctd.py module in the descriptors directory.

        Returns
        -------
        :distribution : pd.Dataframe
            pandas Dataframe of D_CTD values for protein sequence. Output will
            be of the shape N x 15, where 15 is the number of features calculated from
            the descriptor.
        """
        print('\nGetting Distribution (CTD) Descriptors...')
        print('#########################################\n')

        #if attribute already calculated & not empty then return it
        if not self.distribution.empty:
            return self.distribution

        #get descriptor-specific parameters from config file
        property = self.descr_parameters[0]["distribution"][0]["property"]

        #initialise dataframe
        distribution_df = pd.DataFrame()

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            distribution_seq = ctd_distribution(seq, property=property)
            distribution_df = pd.concat([distribution_df, distribution_seq])

        self.distribution = distribution_df

        return self.distribution

    def get_ctd(self):
        """
        Calculate CTD of protein sequences using the respective function in 
        the ctd.py module in the descriptors directory.

        Returns
        -------
        :ctd : pd.Series
            pandas Series of CTD values for protein sequence. Output will
            be of the shape N x 147, where 147 is the number of features calculated from
            the descriptor.
        """
        print('\nGetting CTD Descriptors...')
        print('##########################\n')

        #if attribute already calculated & not empty then return it
        if not self.ctd.empty:
            return self.ctd

        #get descriptor-specific parameters from config file
        property = self.descr_parameters[0]["ctd"][0]["property"]
        all_ctd = self.descr_parameters[0]["ctd"][0]["all"]

        #initialise dataframe
        ctd_df = pd.DataFrame()

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            ctd_seq = ctd_(seq, property=property, all_ctd=all_ctd)
            ctd_df = pd.concat([ctd_df, ctd_seq])

        self.ctd = ctd_df

        return self.ctd

    def get_conjoint_triad(self):
        """
        Calculate Conjoint Triad (CTriad) of protein sequences using the respective function in 
        the conjointTriad.py module in the descriptors directory.

        Returns
        -------
        :conjoint_triad : pd.Dataframe
            pandas Dataframe of CTriad descriptor values for all protein sequences. Dataframe
            will be of the shape N x 343, where 343 is the number of features calculated 
            from the descriptor for a sequence.
        """
        print('\nGetting Conjoint Triad Descriptors...')
        print('#####################################\n')

        #if attribute already calculated & not empty then return it
        if not self.conjoint_triad.empty:
            return self.conjoint_triad

        #initialise dataframe
        conjoint_triad_df = pd.DataFrame()

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            conjoint_triad_seq = conjoint_triad(seq)
            conjoint_triad_df = pd.concat([conjoint_triad_df, conjoint_triad_seq])

        self.conjoint_triad = conjoint_triad_df

        return self.conjoint_triad

    def get_seq_order_coupling_number(self):
        """
        Calculate Sequence Order Coupling number (SOCNum) of protein sequences 
        using the respective function in the quasiSequenceOrder.py module in the descriptors directory.

        Returns
        -------
        :seq_order_df : pd.Dataframe
            Dataframe of SOCNum descriptor values for all protein sequences. Output
            will be of the shape N x M, where N is the number of protein sequences and
            M is the number of features calculated from the descriptor (calculated as
            N * 2 where N = lag).
        """
        print('\nGetting Sequence Order Coupling Descriptors...')
        print('##############################################\n')

        #if attribute already calculated & not empty then return it
        if not self.seq_order_coupling_number.empty:
            return self.seq_order_coupling_number

        #initialise dataframe
        seq_order_coupling_number_df = pd.DataFrame()

        #get descriptor-specific parameters from config file
        lag = self.descr_parameters[0]["seq_order_coupling_number"][0]["lag"]
        distance_matrix = self.descr_parameters[0]["seq_order_coupling_number"][0]["distance_matrix"]

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            seq_order_coupling_number_seq = seq_order_coupling_number(seq, lag=lag, distance_matrix=distance_matrix)
            seq_order_coupling_number_df = pd.concat([seq_order_coupling_number_df, seq_order_coupling_number_seq])

        self.seq_order_coupling_number = seq_order_coupling_number_df

        return self.seq_order_coupling_number

    def get_quasi_seq_order(self):
        """
        Calculate Quasi Sequence Order Coupling number (QSOrder) of protein sequences 
        using the respective function in the quasiSequenceOrder.py module in the descriptors directory.

        Returns
        -------
        :quasi_seq_order_df : pd.Dataframe
            Dataframe of quasi-sequence-order descriptor values for the
            protein sequences, with output shape N x 100 where N is the number
            of sequences and 100 the number of calculated features.
        """
        print('\nGetting Quasi Sequence Order Descriptors...')
        print('###########################################\n')

        #if attribute already calculated & not empty then return it
        if not self.quasi_seq_order.empty:
            return self.quasi_seq_order

        #initialise dataframe
        quasi_seq_order_df = pd.DataFrame()

        #get descriptor-specific parameters from config file
        lag = self.descr_parameters[0]["quasi_seq_order"][0]["lag"]
        weight = self.descr_parameters[0]["quasi_seq_order"][0]["weight"]
        distance_matrix = self.descr_parameters[0]["quasi_seq_order"][0]["distance_matrix"]

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            quasi_seq_order_seq = seq_order_coupling_number(seq, lag=lag, distance_matrix=distance_matrix)
            quasi_seq_order_df = pd.concat([quasi_seq_order_df, quasi_seq_order_seq])

        self.quasi_seq_order = quasi_seq_order_df

        return self.quasi_seq_order

    def get_pseudo_aa_composition(self):
        """
        Calculate Pseudo Amino Acid Composition (PAAComp) of protein sequences 
        using the respective function in the composition.py module in the descriptors directory.

        Returns
        -------
        :pseudo_aa_composition : pd.Dataframe
            Dataframe of pseudo amino acid composition descriptor values
            for the protein sequences of output shape N x 50.
        """
        print('\nGetting Pseudo Amino Acid Composition Descriptors...')
        print('####################################################\n')

        #if attribute already calculated & not empty then return it
        if not self.pseudo_aa_composition.empty:
            return self.pseudo_aa_composition

        #initialise dataframe
        pseudo_aacomp_df = pd.DataFrame()

        #get descriptor-specific parameters from config file
        lamda = self.descr_parameters[0]["pseudo_aa_composition"][0]["lambda"]
        weight = self.descr_parameters[0]["pseudo_aa_composition"][0]["weight"]
        properties = self.descr_parameters[0]["pseudo_aa_composition"][0]["properties"]

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            pseudo_aacomp_seq = pseudoAAC(seq, lamda=lamda, weight=weight, properties=properties)
            pseudo_aacomp_df = pd.concat([pseudo_aacomp_df, pseudo_aacomp_seq])

        self.pseudo_aa_composition = pseudo_aacomp_df

        return self.pseudo_aa_composition
        
    def get_amp_pseudo_aa_composition(self):
        """
        Calculate Amphiphilic Pseudo Amino Acid Composition (APAAComp) of protein sequences 
        using the respective function in the composition.py module in the descriptors directory.

        Returns
        -------
        :amp_pseudo_aa_composition_df : pd.Dataframe
            Dataframe of Amphiphilic pseudo amino acid composition descriptor
            values for the protein sequences of output shape N x 80.
        """
        print('\nGetting Amphiphilic Pseudo Amino Acid Composition Descriptors...')
        print('###############################################################\n')

        #if attribute already calculated & not empty then return it
        if not self.amp_pseudo_aa_composition.empty:
            return self.amp_pseudo_aa_composition

        #get descriptor-specific parameters from config file
        lamda = self.parameters["amp_pseudo_aa_composition"]["lambda"]
        weight = self.parameters["amp_pseudo_aa_composition"]["weight"]

        #calculate descriptor value, concatenate descriptor values
        for seq in self.protein_seqs:
            amp_pseudo_aacomp_seq = amphiphilicPseudoAAC(seq, lamda=lamda, weight=weight)
            amp_pseudo_aacomp_df = pd.concat([amp_pseudo_aacomp_df, amp_pseudo_aacomp_seq])

        self.amp_pseudo_aa_composition = amp_pseudo_aacomp_df

        return self.amp_pseudo_aa_composition
    
    def get_descriptor_encoding(self, descriptor):
        """
        Get the protein descriptor values of a specified input descriptor. If the
        sought descriptor has already been calculated then its attribute is returned,
        else the descriptor is calculated using its get_descriptor function.

        Parameters
        ----------
        :descriptor : str
            name of descriptor to return. Method can accept the approximate name
            of the descriptor, e.g. 'aa_comp'/'aa_compo' etc will return the
            'aa_composition' descriptor. This functionality is realised using
            the difflib library and its built-in get_close_matches function.

        Returns
        -------
        :desc_encoding : pd.DataFrame/None
            dataframe of descriptor values, calculated from the descriptor
            instances' protein sequences. None returned if no matching descriptor found.
        """
        #remove any detected whitespace from input parameter
        try:
          descriptor = descriptor.strip()
        except:
          raise TypeError('Input parameter {} is not of correct datatype string, got {}' \
            .format(descriptor, type(descriptor)))

        #validate input descriptor is a valid available descriptor, get its closest match
        desc_matches = get_close_matches(descriptor,self.valid_descriptors(),cutoff=0.4)
        if desc_matches!=[]:
            desc = desc_matches[0]  #set desc to closest descriptor match found
        else:
            raise ValueError('Could not find a match for the input descriptor {} \
                in available valid models:\n {}.'.format(descriptor, self.valid_descriptors()))

        #if sought descriptor attribute dataframe is empty, call the descriptor's
        #   get_descriptor() function, set desc_encoding to descriptor attribute
        if desc == 'aa_composition':
            if (getattr(self, desc).empty):
                self.get_aa_composition()
            desc_encoding = self.aa_composition
        elif desc == 'dipeptide_composition':
            if (getattr(self, desc).empty):
                self.get_dipeptide_composition()
            desc_encoding = self.dipeptide_composition
        elif desc == 'tripeptide_composition':
            if (getattr(self, desc).empty):
                self.get_tripeptide_composition()
            desc_encoding = self.tripeptide_composition
        elif desc == 'normalized_moreaubroto_autocorrelation':
            if (getattr(self, desc).empty):
              self.get_norm_moreaubroto_autocorrelation()
            desc_encoding = self.normalized_moreaubroto_autocorrelation
        elif desc == 'moran_autocorrelation':
            if (getattr(self, desc).empty):
              self.get_moran_autocorrelation()
            desc_encoding = self.moran_autocorrelation
        elif desc == 'geary_autocorrelation':
            if (getattr(self, desc).empty):
              self.get_geary_autocorrelation()
            desc_encoding = self.geary_autocorrelation
        elif desc == 'ctd':
            if (getattr(self, desc).empty):
              self.get_ctd()
            desc_encoding = self.ctd
        elif desc == 'comp':
            if (getattr(self, desc).empty):
              self.get_composition()
            desc_encoding = self.comp
        elif desc == 'transition':
            if (getattr(self, desc).empty):
              self.get_transition()
            desc_encoding = self.transition
        elif desc == 'distribution':
            if (getattr(self, desc).empty):
              self.get_distribution()
            desc_encoding = self.distribution
        elif desc == 'conjoint_triad':
            if (getattr(self, desc).empty):
              self.get_conjoint_triad()
            desc_encoding = self.conjoint_triad
        elif desc == 'seq_order_coupling_number':
            if (getattr(self, desc).empty):
              self.get_seq_order_coupling_number()
            desc_encoding = self.seq_order_coupling_number
        elif desc == 'quasi_seq_order':
            if (getattr(self, desc).empty):
              self.get_quasi_seq_order()
            desc_encoding = self.quasi_seq_order
        elif desc == 'pseudo_aa_composition':
            if (getattr(self, desc).empty):
              self.get_pseudo_aa_composition()
            desc_encoding = self.pseudo_aa_composition
        elif desc == 'amp_pseudo_aa_composition':
            if (getattr(self, desc).empty):
              self.get_amp_pseudo_aa_composition()
            desc_encoding = self.amp_pseudo_aa_composition
        else:
          desc_encoding = None           #no matching descriptor

        return desc_encoding

    def all_descriptors_list(self, desc_combo=1):
       """
       Get list of all available descriptor attributes. Using the desc_combo
       input parameter you can get the list of all descriptors, all combinations
       of 2 descriptors or all combinations of 3 descriptors. Default of 1 will
       mean a list of all available descriptor attributes will be returned.

       Parameters
       ----------
       :desc_combo : int (default=1)
            combination of descriptors to return. A value of 2 or 3 will return
            all combinations of 2 or 3 descriptor attributes etc.

       Returns
       -------
       :all_descriptors : list
            list of available descriptor attributes.
       """
       #filter out class attributes that are not any of the desired descriptors
       all_descriptors = list(filter(lambda x: x.startswith('_'), list(self.__dict__.keys())))
       all_descriptors = list(filter(lambda x: not x.startswith('_all_desc'), all_descriptors))
       all_descriptors = [de[1:] for de in all_descriptors]

       #get all combinations of 2 or 3 descriptors.
       if desc_combo == 2:
           all_descriptors = list(itertools.combinations(all_descriptors, 2))
       elif desc_combo == 3:
           all_descriptors = list(itertools.combinations(all_descriptors, 3))
       else:
           pass     #if desc_combo not equal to 2 or 3 then use default all_descriptors

       return all_descriptors

    def get_all_descriptors(self):
        """
        Calculate all individual descriptor values, concatenating each descriptor
        Dataframe into one storing all descriptors.

        Returns
        -------
        :all_desc_df : pd.DataFrame
            concatenated dataframe of all individual descriptors. Using the default
            attributes and their associated values, the output will be of the shape
            N x 9920. The 2nd dimension is dependant on several additional parameters
            of some descriptors, including the number of properties and max lag for the
            Autocorrelation descriptors, the max lag for the SOCNum, the number of properties
            and lamda for PAAComp and the lambda for APAAComp. 9920 features are calculated 
            in total when using the default parameters of each descriptor and N represents 
            the number of sequences.
        """
        print('Calculating all descriptor values....\n')
        print('#####################################\n')

        #if descriptor attribute DF is empty then call get_descriptor function
        if (getattr(self, "aa_composition").empty):
                self.aa_composition = self.get_aa_composition()

        if (getattr(self, "dipeptide_composition").empty):
                self.dipeptide_composition = self.get_dipeptide_composition()

        if (getattr(self, "tripeptide_composition").empty):
                self.tripeptide_composition = self.get_tripeptide_composition()

        if (getattr(self, "normalized_moreaubroto_autocorrelation").empty):
            self.normalized_moreaubroto_autocorrelation = self.get_norm_moreaubroto_autocorrelation()

        if (getattr(self, "moran_autocorrelation").empty):
            self.moran_autocorrelation = self.get_moran_autocorrelation()

        if (getattr(self, "geary_autocorrelation").empty):
            self.geary_autocorrelation = self.get_geary_autocorrelation()

        if (getattr(self, "ctd").empty):
                self.ctd = self.get_ctd()

        if (getattr(self, "comp").empty):
                self.ctd = self.get_composition()

        if (getattr(self, "transition").empty):
            self.ctd = self.get_transition()
        
        if (getattr(self, "distribution").empty):
            self.ctd = self.get_distribution()

        if (getattr(self, "conjoint_triad").empty):
                self.conjoint_triad = self.get_conjoint_triad()

        if (getattr(self, "seq_order_coupling_number").empty):
                self.seq_order_coupling_number = self.get_seq_order_coupling_number()

        if (getattr(self, "quasi_seq_order").empty):
                self.quasi_seq_order = self.get_quasi_seq_order()

        if (getattr(self, "pseudo_aa_composition").empty):
                self.pseudo_aa_composition = self.get_pseudo_aa_composition()

        # if (getattr(self, "amp_pseudo_aa_composition").empty):
        #         self.amp_pseudo_aa_composition = self.get_amp_pseudo_aa_composition()

        #append all calculated descriptors to list
        all_desc = [
            self.aa_composition, self.dipeptide_composition, self.tripeptide_composition,
            self.normalized_moreaubroto_autocorrelation, self.moran_autocorrelation,
            self.geary_autocorrelation, self.comp, self.transition,
            self.distribution, self.ctd, self.conjoint_triad, self.seq_order_coupling_number,
            self.quasi_seq_order, self.pseudo_aa_composition
            ]
        # self.amp_pseudo_aa_composition
        #concatenate individual descriptor dataframe attributes
        all_desc_df = pd.concat(all_desc, axis = 1)
        self.all_descriptors = all_desc_df     

        return all_desc_df

    def valid_descriptors(self):
        """
        Get a list of all valid descriptors available in the Descriptors class.

        Returns
        -------
        :valid_desc : list
            list of all valid descriptors that the class supports.
        """
        valid_desc = [
            'aa_composition', 'dipeptide_composition', 'tripeptide_composition',
            'normalized_moreaubroto_autocorrelation','moran_autocorrelation','geary_autocorrelation',
            'ctd', 'comp', 'transition', 'distribution', 'conjoint_triad',
            'seq_order_coupling_number','quasi_seq_order',
            'pseudo_aa_composition', 'amp_pseudo_aa_composition'
        ]
        return valid_desc

######################          Getters & Setters          ######################

    @property
    def all_desc(self):
        return self._all_desc

    @all_desc.setter
    def all_desc(self, val):
        self._all_desc = val

    @property
    def aa_composition(self):
        return self._aa_composition

    @aa_composition.setter
    def aa_composition(self, val):
        self._aa_composition = val

    @property
    def dipeptide_composition(self):
        return self._dipeptide_composition

    @dipeptide_composition.setter
    def dipeptide_composition(self, val):
        self._dipeptide_composition = val

    @property
    def tripeptide_composition(self):
        return self._tripeptide_composition

    @tripeptide_composition.setter
    def tripeptide_composition(self, val):
        self._tripeptide_composition = val

    @property
    def normalized_moreaubroto_autocorrelation(self):
        return self._normalized_moreaubroto_autocorrelation

    @normalized_moreaubroto_autocorrelation.setter
    def normalized_moreaubroto_autocorrelation(self, val):
        self._normalized_moreaubroto_autocorrelation = val

    @property
    def moran_autocorrelation(self):
        return self._moran_autocorrelation

    @moran_autocorrelation.setter
    def moran_autocorrelation(self, val):
        self._moran_autocorrelation = val

    @property
    def geary_autocorrelation(self):
        return self._geary_autocorrelation

    @geary_autocorrelation.setter
    def geary_autocorrelation(self, val):
        self._geary_autocorrelation = val

    @property
    def ctd(self):
        return self._ctd

    @ctd.setter
    def ctd(self, val):
        self._ctd = val

    @property
    def comp(self):
        return self._comp

    @comp.setter
    def comp(self, val):
        self._comp = val

    @property
    def transition(self):
        return self._transition

    @transition.setter
    def transition(self, val):
        self._transition = val

    @property
    def distribution(self):
        return self._distribution

    @distribution.setter
    def distribution(self, val):
        self._distribution = val

    @property
    def conjoint_triad(self):
        return self._conjoint_triad

    @conjoint_triad.setter
    def conjoint_triad(self, val):
        self._conjoint_triad = val

    @property
    def seq_order_coupling_number(self):
        return self._seq_order_coupling_number

    @seq_order_coupling_number.setter
    def seq_order_coupling_number(self, val):
        self._seq_order_coupling_number = val

    @property
    def quasi_seq_order(self):
        return self._quasi_seq_order

    @quasi_seq_order.setter
    def quasi_seq_order(self, val):
        self._quasi_seq_order = val

    @property
    def pseudo_aa_composition(self):
        return self._pseudo_aa_composition

    @pseudo_aa_composition.setter
    def pseudo_aa_composition(self, val):
        self._pseudo_aa_composition = val

    @property
    def amp_pseudo_aa_composition(self):
        return self._amp_pseudo_aa_composition

    @amp_pseudo_aa_composition.setter
    def amp_pseudo_aa_composition(self, val):
        self._amp_pseudo_aa_composition = val

    @property
    def all_descriptors(self):
        return self._all_descriptors

    @all_descriptors.setter
    def all_descriptors(self, val):
        self._all_descriptors = val

    @all_descriptors.deleter
    def all_descriptors(self):
        """ Delete all descriptor attribute dataframes """
        del self._all_descriptors
        del self._aa_composition
        del self._dipeptide_composition
        del self._tripeptide_composition
        del self._normalized_moreaubroto_autocorrelation
        del self._moran_autocorrelation
        del self._geary_autocorrelation
        del self._ctd
        del self._transition
        del self._comp
        del self._distribution
        del self._conjoint_triad
        del self._seq_order_coupling_number
        del self._quasi_seq_order
        del self._pseudo_aa_composition
        del self._amp_pseudo_aa_composition

################################################################################

    def __str__(self):
        return "Descriptor(Num Sequences: {}, Using All Descriptors: {})".format(
            self.num_seqs, self.all_desc)

    def __repr__(self):
        return ('<Descriptor: {}>'.format(self))

    def __len__(self):
        return len(self.all_descriptors)

    def __shape__(self):
        return self.all_descriptors.shape

    # Methods
    # -------
    # import_descriptors(descriptor_file)
    # get_aa_composition()
    # get_dipeptide_composition()
    # get_tripeptide_composition()
    # get_norm_moreaubroto_autocorrelation()
    # get_moran_autocorrelation()
    # get_geary_autocorrelation()
    # ......



__author__ = 'Adam McKenna, https://github.com/amckenna41'
__authorEmail__ = 'amckenna41@qub.ac.uk'
__url__ = 'https://github.com/amckenna41/pySAR'
__credits__ = ['Adam McKenna']
__version__ = "2.0.4"
__license__ = 'MIT'



################################################################################
#################             Model Module Tests               #################
################################################################################

import unittest
import sklearn
import numpy as np

#### Suppress Sklearn warnings ####
def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn
warnings.filterwarnings("ignore",category=DeprecationWarning)
from sklearn.cross_decomposition import PLSRegression
from sklearn.ensemble import RandomForestRegressor, AdaBoostRegressor, BaggingRegressor
from sklearn.linear_model import Lasso
from sklearn.svm import SVR

from pySAR.model import *
class ModelTests(unittest.TestCase):

    def setUp(self):
        """ Create dummy data. """
        self.dummy_X = np.random.ranf(size=100)
        self.dummy_X_2 = np.random.ranf(size=50)
        self.dummy_Y = np.random.randint(10, size=100)
        self.dummy_Y_2 = np.random.randint(20, size=50)

    def test_model(self):
        """ Test Case to check each model type & its associated parameters & attributes. """
        test_models = ['PLSRegression','RandomForestRegressor','AdaBoostRegressor',\
                            'BaggingRegressor','DecisionTreeRegressor','LinearRegression',\
                            'Lasso','SVR','KNeighborsRegressor', 'GradientBoostingRegressor', 'Ridge']

        #iterate through all available algorithms and test them
        for test_mod in range(0, len(test_models)):

            model = Model(test_models[test_mod])
#1.)
            #checking model object is of the correct sklearn model datatype
            self.assertEqual(type(model.model).__name__, test_models[test_mod],
                'Model type is not correct, wanted {}, got {} '.format(
                    test_models[test_mod], type(model.model).__name__
                ))
#2.)        #assert that model has not been fitted
            self.assertFalse(model.model_fitted(), 'Model should not be fitted \
                on initialisation')
#3.)        #verify that parameters input param = {} meaning the default params for the model are used
            self.assertEqual(model.parameters,{},
                'Default Parameters attribute should be an empty dict, but got {}'.format(model.parameters))
#4.)        #verify test split attribute is = 0.2, its default value
            self.assertEqual(model.test_split, 0.2,
                'Default test split attribute should be 0.2, but got {}'.format(model.test_split))
#5.)        #verify that input model type is a valid model for the class
            self.assertTrue(model.algorithm in [item.lower() for item in model.valid_models],
                'Input algorithm {} not in available algorithms: {}'.format(model.algorithm, model.valid_models))
#6.)        #verify repr representation of model object is correct
            self.assertEqual(repr(model), test_models[test_mod],
                'Repr function should return {}, but got {}'.format(test_models[test_mod], repr(model)))
#7.)        #verify algorithm is a regression
            self.assertTrue(sklearn.base.is_regressor(model.model),
                'Model type should be a sklearn regressor.')
#8.)        #fit model and assert it has been fitted
            model.train_test_split(self.dummy_X, self.dummy_Y)
            model.fit()
            self.assertTrue(model.model_fitted(), 'Model has not been fitted')

    def test_model_input_closeness(self):
        """ Test case for testing the algorithm closeness function used to get the
            closest available algorithm to the algorithm input into the class. """
#1.)
        model = Model('plsreg')
        self.assertEqual(model.algorithm, "plsregression")
        self.assertEqual(repr(model), "PLSRegression")

        model = Model('randomfor')
        self.assertEqual(model.algorithm, "randomforestregressor")
        self.assertEqual(repr(model), "RandomForestRegressor")

        model = Model('adaboo')
        self.assertEqual(model.algorithm, "adaboostregressor")
        self.assertEqual(repr(model), "AdaBoostRegressor")

        model = Model('bagging')
        self.assertEqual(model.algorithm, "baggingregressor")
        self.assertEqual(repr(model), "BaggingRegressor")

        model = Model('decisiontree')
        self.assertEqual(model.algorithm, "decisiontreeregressor")
        self.assertEqual(repr(model), "DecisionTreeRegressor")

        model = Model('linear')
        self.assertEqual(model.algorithm, "linearregression")
        self.assertEqual(repr(model), "LinearRegression")

        model = Model('lass')
        self.assertEqual(model.algorithm, "lasso")
        self.assertEqual(repr(model), "Lasso")

        model = Model('kneighbors')
        self.assertEqual(model.algorithm, "kneighborsregressor")
        self.assertEqual(repr(model), "KNeighborsRegressor")

        model = Model('sv')
        self.assertEqual(model.algorithm, "svr")
        self.assertEqual(repr(model), "SVR")

        model = Model('rid')
        self.assertEqual(model.algorithm, "ridge")
        self.assertEqual(repr(model), "Ridge")

        model = Model('gbr')
        self.assertEqual(model.algorithm, "gbr")
        self.assertEqual(repr(model), "GradientBoostingRegressor")

        model = Model('sg')
        self.assertEqual(model.algorithm, "sgd")
        self.assertEqual(repr(model), "SGDRegressor")
#2.)
        with self.assertRaises(ValueError, msg='Value Error raised, invalid model/algorithm name given.'):
            bad_model = Model('abcdefg')

        with self.assertRaises(ValueError, msg='Value Error raised, invalid model/algorithm name given.'):
            bad_model = Model('notamodel')

        with self.assertRaises(ValueError, msg='Value Error raised, invalid model/algorithm name given.'):
            bad_model = Model('123')

        with self.assertRaises(ValueError, msg='Value Error raised, invalid model/algorithm name given.'):
            bad_model = Model('blahblahblah')

    def test_train_test_split(self):
        """ Testing splitting up dataset into training and test data. """
#1.)
        model = Model('plsreg')
        X_train, X_test, Y_train, Y_test = model.train_test_split(self.dummy_X, self.dummy_Y)

        self.assertTrue(len(X_train) == 80)
        self.assertTrue(len(Y_train) == 80)
        self.assertTrue(len(X_test) == 20)
        self.assertTrue(len(Y_test) == 20)
#2.)
        model = Model('plsreg')
        with self.assertRaises(ValueError, msg='Value Error raised, invalid X and Y data given.'):
            X_train, X_test, Y_train, Y_test = model.train_test_split(self.dummy_X_2, self.dummy_Y)
#3.)
        model = Model('adaboostreg')
        X_train, X_test, Y_train, Y_test = model.train_test_split(self.dummy_X, self.dummy_Y, test_size=0.5)

        self.assertTrue(len(X_train) == 50)
        self.assertTrue(len(Y_train) == 50)
        self.assertTrue(len(X_test) == 50)
        self.assertTrue(len(Y_test) == 50)
#4.)
        model = Model('bagging')
        X_train, X_test, Y_train, Y_test = model.train_test_split(self.dummy_X_2, self.dummy_Y_2, test_size=0.1)

        self.assertTrue(len(X_train) == 45)
        self.assertTrue(len(Y_train) == 45)
        self.assertTrue(len(X_test) == 5)
        self.assertTrue(len(Y_test) == 5)

    def test_predict(self):
        """ Testing the prediction of values for unseen sequences using the trained model. """
#1.)
        model = Model('knn')
        X_train, X_test, Y_train, Y_test = model.train_test_split(self.dummy_X_2, self.dummy_Y_2)
        model.fit()
#2.)
        Y_pred = model.predict()
        self.assertIsInstance(Y_pred, np.ndarray)
        self.assertEqual(len(Y_pred), len(Y_test))

    def test_parameters(self):
        """ Testing parameters of Model class. """
#1.)
        #create instance of PLS model using Model class & creating instance
        #   using SKlearn libary, comparing if the parameters of both instances are equal
        pls_parameters = {"n_components": 20, "scale": False, "max_iter": 200}
        model = Model(algorithm="PlsRegression", parameters=pls_parameters)
        pls_model = PLSRegression(n_components=20, scale="svd", max_iter=200)

        for k, v in model.model.get_params().items():
            self.assertIn(k, list(pls_model.get_params()))
#2.)
        rf_parameters = {"n_estimators": 200, "max_depth": 50,"min_samples_split": 10}
        model = Model(algorithm="RandomForest", parameters=rf_parameters)
        rf_model = RandomForestRegressor(n_estimators=200, max_depth=50, min_samples_split=10)

        for k, v in model.model.get_params().items():
            self.assertIn(k, list(rf_model.get_params()))
#3.)
        knn_parameters = {"n_neighbors": 10, "weights": "distance", "algorithm": "ball_tree"}
        model = Model(algorithm="KNN", parameters=knn_parameters)
        knn_model = KNeighborsRegressor(n_neighbors=10, weights='distance', algorithm="kd_tree")

        for k, v in model.model.get_params().items():
            self.assertIn(k, list(knn_model.get_params()))
#4.)
        svr_parameters = {"kernel": "poly", "degree": 5, "coef0": 1}
        model = Model(algorithm="SVR",parameters=svr_parameters)
        svr_model = SVR(kernel='poly', degree=5, coef0=1)

        for k, v in model.model.get_params().items():
            self.assertIn(k, list(svr_model.get_params()))
#5.)
        ada_parameters = {"n_estimators": 150, "learning_rate": 1.2, "loss": "square"}
        model = Model(algorithm="AdaBoost", parameters=ada_parameters)
        ada_model = AdaBoostRegressor(n_estimators=150, learning_rate=1.2, loss="square")

        for k, v in model.model.get_params().items():
            self.assertIn(k, list(ada_model.get_params()))
#6.)
        bagging_parameters = {"n_estimators": 50, "max_samples": 1.5, "max_features": 2}
        model = Model(algorithm="Bagging", parameters=bagging_parameters)
        bagging_model = BaggingRegressor(n_estimators=50, max_samples=1.5, max_features="square")

        for k, v in model.model.get_params().items():
            self.assertIn(k, list(bagging_model.get_params()))
#7.)
        lasso_parameters = {"alpha": 1.5, "max_iter": 500, "tol": 0.004}
        model = Model(algorithm="lasso", parameters=lasso_parameters)
        lasso_model = Lasso(alpha=1.5, max_iter=500, tol=0.004)

        for k, v in model.model.get_params().items():
            self.assertIn(k, list(lasso_model.get_params()))

    def test_copy(self):
        """ Testing model copy function. """
#1.)
        pls_parameters = {"n_components": 20, "scale": False, "max_iter":200}
        model = Model(algorithm="PLSReg", parameters=pls_parameters)
        model_copy = model.copy()

        self.assertTrue(model_copy == model.model)
#2.)
        svr_parameters = {"kernel": "poly", "degree": 5, "coef0": 1}
        model = Model(algorithm="SVR", parameters=svr_parameters)
        model_copy = model.copy()

        self.assertTrue(model_copy == model.model)

    def test_hyperparamter_tuning(self):
        """ Testing hyperparamter tuning function. """
#1.)
        model = Model(algorithm="PLSReg")
        with self.assertRaises(TypeError, msg='Type Error raised, invalid parameter data type given.'):
            model.hyperparameter_tuning(parameters='wrongType')

        with self.assertRaises(UndefinedMetricWarning, msg='Undefined Metric Error raised, invalid hyperparamter tuning metric given.'):
            model.hyperparameter_tuning(metric='blahblah')
#2.)
        #get training and test dataset split
        X_train, X_test, Y_train, Y_test = model.train_test_split(self.dummy_X, self.dummy_Y)

        #fit predictive model and undertake hyperparameter tuning
        model.fit()
        model.hyperparameter_tuning()
    
    def test_feature_selection(self):
        pass
    
    def tearDown(self):
        """ Delete any temp data used for tests. """
        del self.dummy_X
        del self.dummy_X_2
        del self.dummy_Y
        del self.dummy_Y_2

if __name__ == '__main__':
    #run all model tests
    unittest.main(verbosity=2)



################################################################################
#################                  Protein DSP                 #################
################################################################################

from multiprocessing.sharedctypes import Value
import pandas as pd
import numpy as np
from difflib import get_close_matches
from scipy import signal
from scipy.signal import savgol_filter, medfilt, symiirorder1, lfilter, hilbert
from scipy.signal.windows import blackman, hanning, hamming, bartlett, blackmanharris, \
     kaiser, gaussian, barthann, bohman, chebwin, cosine, exponential, boxcar, \
        flattop, nuttall, parzen, tukey, triang
try:
    from scipy.fftpack import fft, ifft, fftfreq, rfft, rfftfreq
except:
    from numpy.fft import fft, ifft, fftfreq, rfft, rfftfreq
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import json
from json import JSONDecodeError

from .utils import *
from .globals_ import DATA_DIR, OUTPUT_DIR, OUTPUT_FOLDER
from .aaindex import AAIndex
class PyDSP():
    """
    Transform protein sequences into their spectral form via a Fast Fourier Transform (FFT).
    Fourier analysis is fundamentally a method for expressing a function as a sum of periodic components,
    and for recovering the function from those components. When both the function and its Fourier
    transform are replaced with discretized counterparts, it is called the Discrete Fourier transform (DFT).
    An implementation algorithm for the DFT is known as the FFT, which is used here. From the FFT
    transformations on the encoded protein sequences, various informational protein spectra
    can be generated, including the power, real, imaginary and absolute spectra.
    Prior to the FFT, a window function can be applied to the sequences which is a mathmatical
    function that applies a weighting to each discrete time series sample in a finite set.
    By default, the hamming function is applied; although the function can accept
    the blackman, blackmanharris, bartlett, gaussian and kaiser window functions. A
    filter function can also be applied. Additionally a convolution function can be 
    applied before the FFT.

    Additonal to the FFT, the RFFT (real-FFT) is calculated which removes some
    redundancy that is generated from the original FFT function which itself is
    Hermitian-symmetric meaning the negative frequency terms are just the complex
    conjugates of the corresponding positive frequency terms, thus the neg-freq
    terms are redundant.

    In the pipeline of pySAR this class and its functions are onyl used when the 'use_dsp'
    parameter is set to true in the config files meaning that the encoded protein sequences
    are passed through a Digital Signal Processing pipeline before being used as training
    data for the regression models. 

    Attributes
    ----------
    :dsp_config (str/json):
        path to configuration file containing DSP parameters OR JSON object of DSP parameters,
        depending on if the parameter is a valid filepath or not.
    :protein_seqs (np.ndarray):
        array of pre-encoded protein sequences. Class accepts only numerically encoded protein
        sequences, not in amino acid form.

    Methods
    -------
    pre_processing():
        complete pre-processing steps before completeing DSP functionality.
    encode_seqs():
        calculate FFT/RFFT of protein seqeuences.
    inverse_FFT():
        calculate inverse FFT of protein sequences.
    consensus_freq():
        calculate consensus frequency of FFT.
    max_freq():
        calculate max frequency of FFT
    """
    def __init__(self, dsp_config="", protein_seqs=None):

        self.protein_seqs = protein_seqs
        self.dsp_config = dsp_config
        self.parameters = {}

        config_filepath = ""
        #read protein seqs from dataset if not input as parameter,
        # parse DSP parameters from config file
        if not isinstance(dsp_config, str) or dsp_config is None:
            raise TypeError('JSON config file must be a filepath of type string, got type {}.'.format(type(dsp_config)))
        if os.path.isfile(self.dsp_config):
            config_filepath = self.dsp_config
        elif os.path.isfile(os.path.join('config', self.dsp_config)):
            config_filepath = os.path.join('config', self.dsp_config)
        else:
            raise OSError('JSON config file not found at path: {}.'.format(config_filepath))
        try:
            with open(config_filepath) as f:
                self.parameters = json.load(f)
        except:
            raise JSONDecodeError('Error parsing config JSON file: {}.'.format(config_filepath))

        #read protein seqs from dataset if not input as parameter
        if (self.protein_seqs is None):
            raise ValueError('Protein sequences input parameter cannot be empty or None.')

        #reshape protein sequences to 2 dimensions
        if (self.protein_seqs.ndim!=2):
            try:
                self.protein_seqs = self.protein_seqs.reshape((-1,1))
            except:
                raise ValueError('Error reshaping input sequences.')

        #set all DSP parameters
        self.dsp_parameters = self.parameters["pyDSP"]
        self.use_dsp = self.dsp_parameters[0]["use_dsp"]
        self.spectrum = self.dsp_parameters[0]["spectrum"]
        self.window_type = self.dsp_parameters[0]["window"]
        self.window = self.dsp_parameters[0]["window"]
        self.filter = self.dsp_parameters[0]["filter"]
        self.convolution = self.dsp_parameters[0]["convolution"]

        #pre-processing of encoded protein sequences
        self.pre_processing()

        #transform sequences into the various informational protein spectra
        self.encode_seqs()

    def pre_processing(self):
        """
        Complete various pre-processing steps for encoded protein sequences before
        doing any of the DSP-related functions or transformations. Zero-pad
        the sequences, remove any +/- infinity or NAN values, get the approximate
        protein spectra and window function parameter names.

        Parameters
        ----------
        :self (PyDSP object): 
            instance of PyDSP class.
            
        Returns
        -------
        None

        """
        #zero-pad encoded sequences so they are all the same length
        self.protein_seqs = zero_padding(self.protein_seqs)

        #get shape parameters of proteins seqs
        self.num_seqs = self.protein_seqs.shape[0]
        self.signal_len = self.protein_seqs.shape[1]

        #replace any positive or negative infinity or NAN values with 0
        self.protein_seqs[self.protein_seqs == -np.inf] = 0
        self.protein_seqs[self.protein_seqs == np.inf] = 0
        self.protein_seqs[self.protein_seqs == np.nan] = 0

        #replace any NAN's with 0's
        #self.protein_seqs.fillna(0, inplace=True)
        self.protein_seqs = np.nan_to_num(self.protein_seqs)

        #initialise zeros array to store all protein spectra
        self.fft_power = np.zeros((self.num_seqs, self.signal_len))
        self.fft_real = np.zeros((self.num_seqs, self.signal_len))
        self.fft_imag = np.zeros((self.num_seqs, self.signal_len))
        self.fft_abs = np.zeros((self.num_seqs, self.signal_len))

        #list of accepted spectra, window functions and filters
        all_spectra = ['power','absolute','real','imaginary']
        all_windows = ['hamming', 'blackman','blackmanharris','gaussian','bartlett',
                       'kaiser', 'barthann', 'bohman', 'chebwin', 'cosine', 'exponential'
                       'flattop', 'hann', 'boxcar', 'hanning', 'nuttall', 'parzen',
                        'triang', 'tukey']
        all_filters = ['savgol', 'medfilt', 'symiirorder1', 'lfilter', 'hilbert']

        #set required input parameters, raise error if spectrum is none
        if self.spectrum == None:
            raise ValueError('Invalid input Spectrum type ({}) not available in valid spectra: {}'.
                format(self.spectrum, all_spectra))
        else:
            #get closest correct spectra from user input, if no close match then raise error
            spectra_matches = (get_close_matches(self.spectrum, all_spectra, cutoff=0.4))

            if spectra_matches == []:
                raise ValueError('Invalid input Spectrum type ({}) not available in valid spectra: {}'.
                    format(self.spectrum, all_spectra))
            else:
                self.spectra = spectra_matches[0]   #closest match in array

        if self.window_type == None:
            self.window = 1       #window = 1 is the same as applying no window
        else:
            #get closest correct window function from user input
            window_matches = (get_close_matches(self.window, all_windows, cutoff=0.4))

            #check if sym=True or sym=False
            #get window function specified by window input parameter, if no match then window = 1
            if window_matches != []:
                if window_matches[0] == 'hamming':
                    self.window = hamming(self.signal_len, sym=True)
                    self.window_type = "hamming"
                elif window_matches[0] == "blackman":
                    self.window = blackman(self.signal_len, sym=True)
                    self.window = "blackman"
                elif window_matches[0] == "blackmanharris":
                    self.window = blackmanharris(self.signal_len, sym=True) #**
                    self.window_type = "blackmanharris"
                elif window_matches[0] == "bartlett":
                    self.window = bartlett(self.signal_len, sym=True)
                    self.window_type = "bartlett"
                elif window_matches[0] == "gaussian":
                    self.window = gaussian(self.signal_len, std=7, sym=True)
                    self.window_type = "gaussian"
                elif window_matches[0] == "kaiser":
                    self.window = kaiser(self.signal_len, beta=14, sym=True)
                    self.window_type = "kaiser"
                elif window_matches[0] == "hanning":
                    self.window = hanning(self.signal_len, sym=True)
                    self.window_type = "hanning"
                elif window_matches[0] == "barthann":
                    self.window = barthann(self.signal_len, sym=True)
                    self.window_type = "barthann"
                elif window_matches[0] == "bohman":
                    self.window = bohman(self.signal_len, sym=True)
                    self.window_type = "bohman"
                elif window_matches[0] == "chebwin":
                    self.window = chebwin(self.signal_len, sym=True)
                    self.window_type = "chebwin"
                elif window_matches[0] == "cosine":
                    self.window = cosine(self.signal_len, sym=True)
                    self.window_type = "cosine"
                elif window_matches[0] == "exponential":
                    self.window = exponential(self.signal_len, sym=True)
                    self.window_type = "exponential"
                elif window_matches[0] == "flattop":
                    self.window = flattop(self.signal_len, sym=True)
                    self.window_type = "flattop"
                elif window_matches[0] == "boxcar":
                    self.window = boxcar(self.signal_len, sym=True)
                    self.window_type = "boxcar"
                elif window_matches[0] == "nuttall":
                    self.window = nuttall(self.signal_len, sym=True)
                    self.window_type = "nuttall"
                elif window_matches[0] == "parzen":
                    self.window = parzen(self.signal_len, sym=True)
                    self.window_type = "parzen"
                elif window_matches[0] == "triang":
                    self.window = triang(self.signal_len, sym=True)
                    self.window_type = "triang"
                elif window_matches[0] == "tukey":
                    self.window = tukey(self.signal_len, sym=True)
                    self.window_type = "tukey"

            else:
                self.window = 1     #window = 1 is the same as applying no window

        #calculate convolution from protein sequences
        if self.convolution is not None:
            if self.window is not None:
                self.convoled_seqs = signal.convolve(self.protein_seqs, self.window, mode='same') / sum(self.window)

        if self.filter != None:
            #get closest correct filter from user input
            filter_matches = (get_close_matches(self.filter, all_filters, cutoff=0.4))

            #set filter attribute according to approximate user input
            if filter_matches != []:
                if filter_matches[0] == 'savgol':
                    self.filter = savgol_filter(self.signal_len, self.signal_len) 
                elif filter_matches[0] == 'medfilt':
                    self.filter = medfilt(self.signal_len)
                elif filter_matches[0] == 'symiirorder1':
                    self.filter = symiirorder1(self.signal_len, c0=1, z1=1)
                elif filter_matches[0] == 'lfilter':
                    self.filter = lfilter(self.signal_len)
                elif filter_matches[0] == 'hilbert':
                    self.filter = hilbert(self.signal_len)      
            else:
                self.filter = ""           #no filter

    def encode_seqs(self): 
        """
        Calculate the FFT and RFFT of the protein sequences already encoded using
        AAI indices, then use the output of the FFT to calculate the various
        informational protein spectra including the power, absolute, real and imaginary.
        The spectrum_encoding attribute will be set to the spectra inputted by
        user from the 'spectrum' config parameter.

        Parameters 
        ----------
        :self (PyDSP object): 
            instance of PyDSP class.

        Returns
        -------
        None
        """
        #create copy of protein sequences so the original instance var remains unchanged
        encoded_seq_copy = np.copy(self.protein_seqs)

        '''From numpy.rfft documentation:
        If n is even, the length of the transformed axis is (n/2)+1.
        If n is odd, the length is (n+1)/2.

        Simply, set the 2nd dimension of the output array size used to store
        rfft output to (N/2)+1 if n is even and (N+1)/2 if odd, where N is the
        signal length (length of sequences).'''
        if (self.signal_len % 2 ==0):
            rfft_output_size = int((self.signal_len/2) + 1)
        else:
            rfft_output_size = int((self.signal_len+1)/2)

        #initialise zero arrays used to store output of both fft and rfft, set
        #  datatype to complex number as that is the output type of the transformation.
        encoded_dataset_rfft = np.zeros((self.protein_seqs.shape), dtype=complex)
        encoded_dataset_fft = np.zeros((self.protein_seqs.shape), dtype=complex)

        #initialise zero arrays used to store output frequencies from fft & rfft transformations
        rttf_freq_size = int((rfft_output_size)/2 + 1)
        encoded_freqs_rfft = np.zeros((self.protein_seqs.shape))
        encoded_freqs_fft = np.zeros(self.protein_seqs.shape)

        #iterate through each sequence, applying the FFT and RFFT to each
        for seq in range(0, self.num_seqs):

          encoded_rfft = np.zeros((self.protein_seqs.shape[1]), dtype=complex)
          encoded_fft = np.zeros((self.protein_seqs.shape[1]), dtype=complex)

          #apply filter *
        #   if (self.filter!=""):
        #      if (self.filter == 'savgol'):
        #         encoded_seq_copy[seq] = savgol_filter(encoded_seq_copy[seq], 17, polyorder=2, deriv=2)

          if not self.convolution:
            #apply window function to Fourier array
            encoded_rfft = rfft(encoded_seq_copy[seq] * self.window)
            encoded_fft = fft(encoded_seq_copy[seq] * self.window)
          else:
            #apply convolution and window function to encoded sequences
            encoded_rfft = rfft((signal.convolve(encoded_seq_copy[seq], self.window, mode='same') / sum(self.window)))
            encoded_fft = fft((signal.convolve(encoded_seq_copy[seq], self.window, mode='same') / sum(self.window)))

          #append transformation from current sequence seq to array of all transformed seqeunces
          encoded_dataset_rfft[seq] = encoded_rfft
          encoded_dataset_fft[seq] = encoded_fft

          #calcualte FFT/RFFT frequencies
          # freqs_rfft = rfftfreq(encoded_rfft.shape[0], self.signal_len)
          freqs_rfft = rfftfreq(encoded_rfft.size)
          freqs_fft = np.fft.fftfreq(encoded_fft.size)

          #append frequency from current sequence seq to array of all frequencies
          encoded_freqs_rfft[seq] = freqs_rfft
          encoded_freqs_fft[seq] = freqs_fft

        #set FFT and RFFT sequences and frequencies
        self.fft = encoded_dataset_fft
        self.rfft = encoded_dataset_rfft
        self.fft_freqs = encoded_freqs_fft
        self.rfft_freqs = encoded_freqs_rfft

        #get individual spectral values, calculated from the FFT and RFFT transformations
        self.fft_abs = abs(self.fft/self.signal_len)
        self.rfft_abs = abs(self.rfft/self.signal_len)
        self.fft_power = np.abs(self.fft[0:len(self.fft)])
        self.rfft_power = np.abs(self.rfft[0:len(self.rfft)])
        self.fft_real = self.fft.real
        self.rfft_real = self.rfft.real
        self.fft_imag = self.fft.imag
        self.rfft_imag = self.rfft.imag

        #set the spectrum_encoding attribute to the spectra specified by 'spectra'
        #   class input parameter.
        if self.spectra == 'power':
            self.spectrum_encoding = self.fft_power
        elif self.spectra == 'real':
            self.spectrum_encoding = self.fft_real
        elif self.spectra == 'imaginary':
            self.spectrum_encoding = self.fft_imag
        elif self.spectra == 'absolute':
            self.spectrum_encoding = self.fft_abs

    def inverse_FFT(self, a, n):
        """
        Get the inverse Fourier Transform of FFT.

        Parameters
        ----------
        :a : np.ndarray
            input array of 1D Fourier Transform.
        :n : int
            length of the output.

        Returns
        -------
        :inv_FFT : np.ndarray
            array of inverse Fourier Transform.
        """
        self.inv_FFT = np.fft.ifft(a,n)
        return self.inv_FFT

    def consensus_freq(self, freqs):
        """
        Get the Consensus Frequency from Fourier Transform of encoded protein sequences.

        Parameters
        ----------
        :freqs : np.ndarray
            frequencies of Fourier Transform.

        Returns
        -------
        :CF : float
            consensus frequency found in array of frequencies.
        :CFi : int
            index of consensus frequency.
        """
        # CF = PP/N ( peak position/length of largest protein in dataset)
        CF, CFi = (self.max_freq(freqs))/self.num_seqs

        return CF, CFi

    def max_freq(self, freqs):
        """
        Get the maximum frequency from Fourier Transform of encoded protein sequences.

        Parameters
        ----------
        :freqs : np.ndarray
            frequencies of Fourier Transform.

        Returns
        -------
        :max_F : float
            maximum frequency found in array of frequencies.
        :max_FI : int
            index of maximum frequency.
        """
        max_F = max(freqs)
        max_FI = np.argmax(freqs)

        return max_F, max_FI

######################          Getters & Setters          ######################

    @property
    def fft_power(self):
        return self._fft_power

    @fft_power.setter
    def fft_power(self, val):
        self._fft_power = val

    @property
    def fft_real(self):
        return self._fft_real

    @fft_real.setter
    def fft_real(self, val):
        self._fft_real = val

    @property
    def fft_imag(self):
        return self._fft_imag

    @fft_imag.setter
    def fft_imag(self, val):
        self._fft_imag = val

    @property
    def fft_abs(self):
        return self._fft_abs

    @fft_abs.setter
    def fft_abs(self, val):
        self._fft_abs = val

    @property
    def fft_freqs(self):
        return self._fft_freqs

    @fft_freqs.setter
    def fft_freqs(self, val):
        self._fft_freqs = val

    @property
    def window(self):
        return self._window

    @window.setter
    def window(self, val):
        self._window = val

    @property
    def window_type(self):
        return self._window_type

    @window_type.setter
    def window_type(self, val):
        self._window_type = val

    @property
    def filter_(self):
        return self._filter

    @filter_.setter
    def filter_(self, val):
        self._filter = val

    @property
    def convolution(self):
        return self._convolution

    @convolution.setter
    def convolution(self, val):
        self._convolution = val

################################################################################

    def __str__(self):
        return "Instance of PyDSP class, using parameters: {}".format(self.__dict__.keys())

    def __repr__(self):
        return ('<PyDSP: {}>'.format(self))



################################################################################
#################            Descriptors Module Tests          #################
################################################################################

import pandas as pd
import numpy as np
import os
import unittest
unittest.TestLoader.sortTestMethodsUsing = None
from json import JSONDecodeError

import pySAR.descriptors_ as descr
import pySAR.descriptors.composition as descriptor_comp
from pySAR.descriptors import *

class DescriptorTests(unittest.TestCase):

    def setUp(self):
        """ Import the 4 config files for each of the 4 datasets used for testing the descriptor methods. """        
        #array of config files for each test dataset
        config_path = os.path.join('tests', 'test_config')
        self.all_config_files = [
            os.path.join(config_path, "test_thermostability.json"), 
            os.path.join(config_path, "test_enantioselectivity.json"),
            os.path.join(config_path, "test_absorption.json"), 
            os.path.join(config_path, "test_localization.json")
        ]

        #array of the total number of protein seqs per dataset
        self.num_seqs = [261, 152, 81, 254]

    def test_descriptor(self):
        """ Test descriptor initialisation process. Verify the initial input parameters
            and descriptor attributes are correct. """
        #testing on all 4 datasets/config files
        for dataset in range(0, len(self.all_config_files)):
            desc = descr.Descriptors(desc_config=self.all_config_files[dataset])
#1.)
            #verify num_seqs descriptors attribute is correct
            self.assertEqual(desc.num_seqs, self.num_seqs[dataset], 'num_seqs attribute not equal to the number of sequences found.')

            #verify that all input sequences dont have any gaps/missing amino acids
            for seq in desc.protein_seqs:
                self.assertNotIn('-', seq, 'There should be no gaps (-) in the sequences.')
#2.)
            #verify all descriptor attributes are initialised to empty dataframes
            self.assertTrue(desc.aa_composition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.dipeptide_composition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.tripeptide_composition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.normalized_moreaubroto_autocorrelation.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.moran_autocorrelation.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.geary_autocorrelation.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.ctd.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.comp.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.transition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.distribution.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.conjoint_triad.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.seq_order_coupling_number.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.quasi_seq_order.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.pseudo_aa_composition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.amp_pseudo_aa_composition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.all_descriptors.empty, 'Attribute should be initialised to an empty dataframe')
#3.)
            #test Value and Type error exceptions are thrown if invalid parameters input
            with self.assertRaises(TypeError, msg='Type Error raised, incorrect datatype input to class.'):
                fail_desc = descr.Descriptors(desc_config=123)

            with self.assertRaises(TypeError, msg='Type Error raised, incorrect datatype input to class.'):
                fail_desc = descr.Descriptors(desc_config=None)

            with self.assertRaises(OSError, msg='OS Error raised, filepath to config file not found.'):
                fail_desc = descr.Descriptors(desc_config="incorrect_filepath.json")

            with self.assertRaises(OSError, msg='OS Error raised, filepath to config file not found.'):
                fail_desc = descr.Descriptors(desc_config="")

    def test_descriptor_groups(self):
        """ Testing the descriptor groups dictionary which stores the specific group
        that a descriptor attribute is a member of. """
        #testing on all 4 datasets
        for dataset in range(0, len(self.all_config_files)):
            desc = descr.Descriptors(desc_config=self.all_config_files[dataset])
#1.)
            self.assertEqual(list(desc.descriptor_groups.keys()), desc.all_descriptors_list())
            self.assertEqual(list(desc.descriptor_groups.values()).count("Composition"), 3)
            self.assertEqual(list(desc.descriptor_groups.values()).count("Autocorrelation"), 3)
            self.assertEqual(list(desc.descriptor_groups.values()).count("Conjoint Triad"), 1)
            self.assertEqual(list(desc.descriptor_groups.values()).count("Quasi-Sequence-Order"), 2)
            self.assertEqual(list(desc.descriptor_groups.values()).count("CTD"), 4)
            self.assertEqual(list(desc.descriptor_groups.values()).count("Pseudo Composition"), 2)
            self.assertEqual(list(desc.descriptor_groups.values()).count("Conjoint Triad"), 1)
            self.assertEqual(len(desc.descriptor_groups.keys()), len(desc.all_descriptors_list()))
#2.)
            self.assertEqual(desc.descriptor_groups['aa_composition'], "Composition")
            self.assertEqual(desc.descriptor_groups['dipeptide_composition'], "Composition")
            self.assertEqual(desc.descriptor_groups['tripeptide_composition'], "Composition")
            self.assertEqual(desc.descriptor_groups['moran_autocorrelation'], "Autocorrelation")
            self.assertEqual(desc.descriptor_groups['geary_autocorrelation'], "Autocorrelation")
            self.assertEqual(desc.descriptor_groups['normalized_moreaubroto_autocorrelation'], "Autocorrelation")
            self.assertEqual(desc.descriptor_groups['ctd'], "CTD")
            self.assertEqual(desc.descriptor_groups['comp'], "CTD")
            self.assertEqual(desc.descriptor_groups['distribution'], "CTD")
            self.assertEqual(desc.descriptor_groups['transition'], "CTD")
            self.assertEqual(desc.descriptor_groups['conjoint_triad'], "Conjoint Triad")
            self.assertEqual(desc.descriptor_groups['pseudo_aa_composition'], "Pseudo Composition")
            self.assertEqual(desc.descriptor_groups['quasi_seq_order'], "Quasi-Sequence-Order")
            self.assertEqual(desc.descriptor_groups['seq_order_coupling_number'], "Quasi-Sequence-Order")
            self.assertEqual(desc.descriptor_groups['amp_pseudo_aa_composition'], "Pseudo Composition")
#3.)
            self.assertIsInstance(desc.descriptor_groups, dict)

    def test_all_descriptors_list(self):
        """ Testing function that returns various combinations of available descriptors 
            using built-in itertools library. """
        desc = descr.Descriptors(desc_config=self.all_config_files[0])
#1.)
        self.assertEqual(len(desc.all_descriptors_list(desc_combo=1)), 15)
        self.assertEqual(len(desc.all_descriptors_list(desc_combo=2)), 105)
        self.assertEqual(len(desc.all_descriptors_list(desc_combo=3)), 455)
#2.)
        self.assertIsInstance(desc.all_descriptors_list(desc_combo=1), list)
        self.assertIsInstance(desc.all_descriptors_list(desc_combo=2), list)
        self.assertIsInstance(desc.all_descriptors_list(desc_combo=3), list)

    def test_valid_descriptors(self):
        """ Testing function that returns the list of valid descriptors available in descriptors module. """
        desc = descr.Descriptors(desc_config=self.all_config_files[0])
        valid_desc = desc.valid_descriptors()
#1.)
        self.assertEqual(len(valid_desc), 15)
        self.assertIsInstance(valid_desc, list)
        self.assertIn('seq_order_coupling_number', valid_desc)
        self.assertIn('geary_autocorrelation', valid_desc)
        self.assertIn('normalized_moreaubroto_autocorrelation', valid_desc)
        self.assertIn('distribution', valid_desc)
        self.assertIn('pseudo_aa_composition', valid_desc)
        self.assertIn('aa_composition', valid_desc)
        self.assertIn('tripeptide_composition', valid_desc)
        self.assertIn('dipeptide_composition', valid_desc)

    def test_aa_composition(self):
        """ Testing Amino Acid Composition protein descriptor attributes and methods. """
        #run tests on all test datasets
        for dataset in range(0, len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])

            #get descriptor values
            aa_comp = desc.get_aa_composition()
#1.)
            self.assertFalse(aa_comp.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(desc.aa_composition.equals(aa_comp), 'Output dataframe and class attribute dataframes must be the same.')
            self.assertEqual(aa_comp.shape, (self.num_seqs[dataset],20), 'Descriptor not of correct shape.') 
            self.assertIsInstance(aa_comp, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertEqual(descriptor_comp.aminoAcids, list(aa_comp.columns), 'Incorrect column values found.')
            self.assertTrue(aa_comp.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

    def test_dipeptide_composition(self):
        """ Testing Dipeptide Composition protein descriptor attributes and methods. """
        #run tests on all test datasets
        for dataset in range(0, len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])

            #get descriptor values
            dipeptide_comp = desc.get_dipeptide_composition()
#1.)
            self.assertTrue(desc.dipeptide_composition.equals(dipeptide_comp), 'Output dataframe and class attribute dataframes must be the same.')
            self.assertFalse(dipeptide_comp.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(dipeptide_comp.shape, (self.num_seqs[dataset],400), 'Descriptor not of correct shape ({},400)'.format(self.num_seqs[dataset]))
            self.assertIsInstance(dipeptide_comp, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertTrue(dipeptide_comp.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

    @unittest.skip("Descriptor can take quite a bit of time to calculate therefore skipping")
    def test_tripeptide_composition(self):
        """ Testing Tripeptide Composition protein descriptor attributes and methods. """
        #run tests on all test datasets
        for dataset in range(0, len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])

            #get descriptor values
            tripeptide_comp = desc.get_tripeptide_composition()
#1.)
            self.assertFalse(tripeptide_comp.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(tripeptide_comp.shape, (self.num_seqs[dataset],8000), 'Descriptor not of correct ({},8000)'.format(self.num_seqs[dataset]))
            self.assertTrue(desc.tripeptide_composition.equals(tripeptide_comp), 'Output dataframe and class attribute dataframes must be the same.')
            self.assertIsInstance(tripeptide_comp, pd.DataFrame, 'Descriptor not of type DataFrame.')
            self.assertTrue(tripeptide_comp.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

    def test_normalized_moreaubroto_autocorrelation(self):
        """ Testing normalized moreaubroto autocorrelation descriptor attributes and methods. """
        #run tests on all test datasets
        for dataset in range(0, len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])

            #get descriptor values
            norm_moreaubroto = desc.get_norm_moreaubroto_autocorrelation()
#1.)
            self.assertFalse(norm_moreaubroto.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(norm_moreaubroto.shape, (self.num_seqs[dataset],240), 'Descriptor not of correct ({},240)'.format(self.num_seqs[dataset]))
            self.assertIsInstance(norm_moreaubroto, pd.DataFrame)
            self.assertTrue(norm_moreaubroto.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

    def test_moran_autocorrelation(self):
        """ Testing Moran autocorrelation descriptor attributes and methods. """
        #run tests on all test datasets
        for dataset in range(0, len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])
            
            #get descriptor values
            moran = desc.get_moran_autocorrelation()
#1.)
            self.assertEqual(moran.shape, (self.num_seqs[dataset],240), 'Descriptor not of correct ({},240)'.format(self.num_seqs[dataset]))
            self.assertIsInstance(moran, pd.DataFrame)
            self.assertFalse(moran.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(moran.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

    def test_geary_autocorrelation(self):
        """ Testing Geary autocorrelation descriptor attributes and methods. """
        #run tests on all test datasets
        for dataset in range(0, len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])

            #get descriptor values
            geary = desc.get_geary_autocorrelation()
#1.)
            self.assertFalse(geary.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(geary.shape, (self.num_seqs[dataset],240), 'Descriptor not of correct ({},240)'.format(self.num_seqs[dataset]))
            self.assertIsInstance(geary, pd.DataFrame)
            self.assertTrue(geary.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

    def test_ctd(self):
        """ Testing CTD descriptor attributes and methods. """
        #run tests on all test datasets
        for dataset in range(0, len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])

            #get descriptor values
            ctd = desc.get_ctd()
#1.)
            self.assertFalse(ctd.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(ctd.shape, (self.num_seqs[dataset], 147), 'Descriptor not of correct ({},147)'.format(self.num_seqs[dataset]))
            self.assertIsInstance(ctd, pd.DataFrame)
            self.assertTrue(ctd.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

            #get descriptor values
            comp = desc.get_composition()
#2.)
            self.assertFalse(comp.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(comp.shape, (self.num_seqs[dataset], 3), 'Descriptor not of correct ({},3)'.format(self.num_seqs[dataset]))
            self.assertIsInstance(comp, pd.DataFrame)
            self.assertTrue(comp.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

            #get descriptor values
            trans = desc.get_transition()
#3.)
            self.assertFalse(trans.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(trans.shape, (self.num_seqs[dataset], 3), 'Descriptor not of correct ({},3)'.format(self.num_seqs[dataset]))
            self.assertIsInstance(trans, pd.DataFrame)
            self.assertTrue(trans.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

            #get descriptor values
            distr = desc.get_distribution()
#4.)
            self.assertFalse(distr.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(distr.shape, (self.num_seqs[dataset],15), 'Descriptor not of correct ({},15)'.format(self.num_seqs[dataset]))
            self.assertIsInstance(distr, pd.DataFrame)
            self.assertTrue(distr.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

    def test_conjoint_triad(self):
        """ Testing Conjoint Triad descriptor attributes and methods. """
        #run tests on all test datasets
        for dataset in range(0, len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])

            #get descriptor values
            conjoint_triad = desc.get_conjoint_triad()
#1.)
            self.assertFalse(conjoint_triad.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(conjoint_triad.shape, (self.num_seqs[dataset],343), 'Descriptor not of correct shape (1,343)')
            self.assertIsInstance(conjoint_triad, pd.DataFrame,'Descriptor not of type DataFrame')
            self.assertTrue(conjoint_triad.any().isnull().sum()==0,'Descriptor should not contain any null values.')

    def test_seq_order_coupling_number(self):
        """ Testing sequence order coupling number descriptor attributes and methods. """
        #run tests on all test datasets
        for dataset in range(0,len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])

            #get descriptor values
            seq_order_coupling_number = desc.get_seq_order_coupling_number()
#1.)
            self.assertFalse(seq_order_coupling_number.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(seq_order_coupling_number.shape, (self.num_seqs[dataset],30), 'Descriptor not of correct shape (1,60)')
            self.assertIsInstance(seq_order_coupling_number, pd.DataFrame, 'Descriptor not of type DataFrame')
            self.assertTrue(seq_order_coupling_number.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

    def test_quasi_seq_order(self):
        """ Testing Quasi sequence order descriptor attributes and methods. """
        #run tests on all test datasets
        for dataset in range(0,len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])

            #get descriptor values
            quasi_seq_order = desc.get_quasi_seq_order()
#1.)
            self.assertFalse(quasi_seq_order.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(quasi_seq_order.shape, (self.num_seqs[dataset],30), 'Descriptor not of correct shape (1,100)')
            self.assertIsInstance(quasi_seq_order, pd.DataFrame, 'Descriptor not of type DataFrame')
            self.assertTrue(quasi_seq_order.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

    @unittest.skip("Descriptor can take quite a bit of time to calculate therefore skipping")
    def test_pseudo_AAC(self):
        """ Testing Pseudo Amino Acid Composition descriptor attributes and methods. """
        #run tests on all test datasets
        for dataset in range(0,len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])

            #get descriptor values
            pseudo_AAC = desc.get_pseudo_aa_composition()
#1.)
            self.assertFalse(pseudo_AAC.empty, 'Descriptor dataframe should not be empty')
            self.assertEqual(pseudo_AAC.shape, (self.num_seqs[dataset],50), 'Descriptor not of correct shape (1,50)')
            self.assertIsInstance(pseudo_AAC, pd.DataFrame, 'Descriptor not of type DataFrame')
            self.assertTrue(pseudo_AAC.any().isnull().sum()==0, 'Descriptor should not contain any null values.')

    @unittest.skip("Descriptor can take quite a bit of time to calculate therefore skipping")
    def test_amp_pseudo_AAC(self):
        """ Testing Amphiphilic Pseudo Amino Acid Composition descriptor attributes and methods. """
        #run tests on all test datasets
        for dataset in range(0,len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])

            #get descriptor values
            amp_pseudo_AAC = desc.get_amp_pseudo_aa_composition()
#1.)
            self.assertFalse(amp_pseudo_AAC.empty)
            self.assertEqual(amp_pseudo_AAC.shape, (self.num_seqs[dataset],80), 'Descriptor not of correct shape (1,80)')
            self.assertIsInstance(amp_pseudo_AAC, pd.DataFrame)
            self.assertTrue(amp_pseudo_AAC.any().isnull().sum()==0,'Descriptor should not contain any null values.')

    @unittest.skip("Test case requires recalculating all descriptors which is redundant to the above tests")
    def test_get_descriptor_encoding(self):
        """ Testing get_descriptor_encoding function. """

        valid_desc = [
            'aa_composition', 'dipeptide_composition', \
            'normalized_moreaubroto_autocorrelation', 'moran_autocorrelation', 'geary_autocorrelation', \
            'ctd', 'comp', 'transition', 'distribution', 'conjoint_triad', \
            'seq_order_coupling_number', 'tripeptide_composition', 'seq_order_coupling_number', \
                'quasi_seq_order', 'pseudo_aa_composition'
        ]
#1.)
        #run tests on all test datasets
        for dataset in range(0,len(self.all_config_files)):
            desc = descr.Descriptors(self.all_config_files[dataset])

            #test each descriptor is calculated when string of its name passed to func
            for val_desc in range(0, len(valid_desc)):

                encoded_desc = desc.get_descriptor_encoding(valid_desc[val_desc])
                self.assertIsNotNone(encoded_desc, 'Descriptor attribute should not be none.')
                self.assertFalse(encoded_desc.empty, 'Descriptor attribute should not be empty.')
                self.assertTrue(encoded_desc.shape[0], len(desc.protein_seqs), 
                    '1st dimension of descriptor attribute should equal num of protein seqs.')

    def tearDown(self):
        """ Cleanup tests and delete datasets/config files. """
        del self.all_config_files
